import json
import logging
import os
import typing
from contextlib import contextmanager
from copy import copy, deepcopy
from enum import Enum

from qtpy.QtCore import Qt, Signal
from qtpy.QtGui import QIcon
from qtpy.QtWidgets import (
    QAction,
    QCheckBox,
    QComboBox,
    QCompleter,
    QDialog,
    QFileDialog,
    QGridLayout,
    QGroupBox,
    QInputDialog,
    QLabel,
    QLineEdit,
    QListWidget,
    QListWidgetItem,
    QMenu,
    QMessageBox,
    QPushButton,
    QSplitter,
    QTabWidget,
    QTextEdit,
    QTreeWidget,
    QTreeWidgetItem,
    QVBoxLayout,
    QWidget,
)
from superqt import QEnumComboBox

from PartSeg._roi_analysis.partseg_settings import PartSettings
from PartSeg._roi_analysis.profile_export import ExportDialog, ImportDialog
from PartSeg.common_backend.except_hook import show_warning
from PartSeg.common_gui.custom_load_dialog import PLoadDialog
from PartSeg.common_gui.custom_save_dialog import FormDialog, PSaveDialog
from PartSeg.common_gui.mask_widget import MaskWidget
from PartSeg.common_gui.searchable_list_widget import SearchableListWidget
from PartSeg.common_gui.universal_gui_part import right_label
from PartSegCore.algorithm_describe_base import AlgorithmProperty, ROIExtractionProfile
from PartSegCore.analysis import SegmentationPipeline
from PartSegCore.analysis.algorithm_description import AnalysisAlgorithmSelection
from PartSegCore.analysis.calculation_plan import (
    CalculationPlan,
    MaskBase,
    MaskCreate,
    MaskFile,
    MaskIntersection,
    MaskSub,
    MaskSuffix,
    MaskSum,
    MeasurementCalculate,
    NodeType,
    PlanChanges,
    RootType,
    Save,
)
from PartSegCore.analysis.measurement_calculation import MeasurementProfile
from PartSegCore.analysis.save_functions import save_dict
from PartSegCore.io_utils import LoadPlanExcel, LoadPlanJson, SaveBase
from PartSegCore.universal_const import Units
from PartSegData import icons_dir

group_sheet = (
    "QGroupBox {border: 1px solid gray; border-radius: 9px; margin-top: 0.5em;} "
    "QGroupBox::title {subcontrol-origin: margin; left: 10px; padding: 0 3px 0 3px;}"
)

MAX_CHANNEL_NUM = 10


class MaskDialog(QDialog):
    def __init__(self, mask_names):
        super().__init__()
        self.mask_names = mask_names
        completer = QCompleter(list(mask_names))
        completer.setCaseSensitivity(Qt.CaseSensitivity.CaseInsensitive)
        self.setWindowTitle("Masks name choose")
        self.mask1_name = QLineEdit()
        self.cancel_btn = QPushButton("Cancel")
        self.ok_btn = QPushButton("Ok")

        self.mask1_name.setCompleter(completer)
        self.mask1_name.textChanged.connect(self.text_changed)
        self.cancel_btn.clicked.connect(self.close)
        self.ok_btn.clicked.connect(self.accept)
        self.ok_btn.setDisabled(True)

        layout = QGridLayout()
        layout.addWidget(right_label("Mask 1 name:"), 0, 0)
        layout.addWidget(self.mask1_name, 0, 1)
        layout.addWidget(self.cancel_btn, 2, 0)
        layout.addWidget(self.ok_btn, 2, 1)
        self.setLayout(layout)

    def text_changed(self):
        text1 = self.get_result()[0]
        if not text1 or text1 not in self.mask_names:
            self.ok_btn.setDisabled(True)
        else:
            self.ok_btn.setDisabled(False)

    def get_result(self):
        text1 = str(self.mask1_name.text()).strip()
        return (text1,)


class TwoMaskDialog(QDialog):
    def __init__(self, mask_names: typing.Iterable[str]):
        """
        :param mask_names: iterable collection of all available mask names
        """
        super().__init__()
        self.mask_names = mask_names
        completer = QCompleter(list(mask_names))
        completer.setCaseSensitivity(Qt.CaseSensitivity.CaseInsensitive)
        self.setWindowTitle("Masks name choose")
        self.mask1_name = QLineEdit()
        self.mask2_name = QLineEdit()
        self.cancel_btn = QPushButton("Cancel")
        self.ok_btn = QPushButton("Ok")

        self.mask1_name.setCompleter(completer)
        self.mask1_name.textChanged.connect(self.text_changed)
        self.mask2_name.setCompleter(completer)
        self.mask2_name.textChanged.connect(self.text_changed)
        self.cancel_btn.clicked.connect(self.close)
        self.ok_btn.clicked.connect(self.accept)
        self.ok_btn.setDisabled(True)

        layout = QGridLayout()
        layout.addWidget(right_label("Mask 1 name:"), 0, 0)
        layout.addWidget(self.mask1_name, 0, 1)
        layout.addWidget(right_label("Mask 2 name:"), 1, 0)
        layout.addWidget(self.mask2_name, 1, 1)
        layout.addWidget(self.cancel_btn, 2, 0)
        layout.addWidget(self.ok_btn, 2, 1)
        self.setLayout(layout)

    def text_changed(self):
        text1, text2 = self.get_result()
        if "" in {text1, text2} or text1 not in self.mask_names or text2 not in self.mask_names:
            self.ok_btn.setDisabled(True)
        else:
            self.ok_btn.setDisabled(text1 == text2)

    def get_result(self):
        text1 = str(self.mask1_name.text()).strip()
        text2 = str(self.mask2_name.text()).strip()
        return text1, text2


class FileMaskType(Enum):
    """
    Enum for file mask types
    """

    Suffix = 0
    Replace = 1
    Mapping_file = 2


class FileMask(QWidget):
    value_changed = Signal()

    def __init__(self):
        super().__init__()
        self.select_type = QEnumComboBox(enum_class=FileMaskType)
        self.values = ["_mask", ("", ""), ""]
        self.first_text = QLineEdit(self.values[0])
        self.second_text = QLineEdit()
        self.first_label = QLabel("Use suffix:")
        self.second_label = QLabel("Replace:")
        self.select_file_btn = QPushButton("Select file")
        self.state = FileMaskType.Suffix

        layout = QGridLayout()
        layout.addWidget(self.select_type, 0, 0, 1, 2)
        layout.addWidget(self.first_label, 1, 0)
        layout.addWidget(self.second_label, 1, 1)
        layout.addWidget(self.first_text, 2, 0)
        layout.addWidget(self.second_text, 2, 1)
        layout.addWidget(self.select_file_btn, 3, 0, 1, 2)
        layout.setColumnStretch(0, 1)
        layout.setRowStretch(4, 1)
        self.setLayout(layout)

        self.second_text.setHidden(True)
        self.second_label.setHidden(True)
        self.select_file_btn.setHidden(True)

        self.first_text.textChanged.connect(self._value_change_wrap)
        self.second_text.textChanged.connect(self._value_change_wrap)
        self.select_type.currentIndexChanged.connect(self._value_change_wrap)
        self.select_type.currentEnumChanged.connect(self.change_type)
        self.select_file_btn.clicked.connect(self.select_file)

    def _value_change_wrap(self, _val=None):
        """Pyside bug workaround"""
        self.value_changed.emit()

    def change_type(self, index: FileMaskType):
        if self.state == index:
            return
        if self.state == FileMaskType.Replace:
            self.values[1] = self.first_text.text(), self.second_text.text()
        else:
            self.values[self.state.value] = self.first_text.text()
        if index == FileMaskType.Replace:
            self.second_text.setHidden(False)
            self.second_label.setHidden(False)
            self.layout().setColumnStretch(1, 1)
            self.first_text.setText(self.values[1][0])
            self.second_text.setText(self.values[1][1])
            self.first_label.setText("Base")
            self.select_file_btn.setHidden(True)
        else:
            self.second_text.setHidden(True)
            self.second_label.setHidden(True)
            self.layout().setColumnStretch(1, 0)
            self.first_label.setText(["Use suffix:", "", "Path:"][index.value])
            self.first_text.setText(self.values[index.value])
            self.select_file_btn.setHidden(index == 0)
        self.state = index

    def select_file(self):
        dial = QFileDialog()
        dial.setFileMode(QFileDialog.ExistingFile)
        dial.setAcceptMode(QFileDialog.AcceptOpen)

        if dial.exec_():
            self.first_text.setText(dial.selectedFiles()[0])

    def is_valid(self) -> bool:
        if self.select_type.currentEnum() == FileMaskType.Suffix:
            return bool(self.first_text.text().strip())
        if self.select_type.currentEnum() == FileMaskType.Replace:
            return "" not in {self.first_text.text().strip(), self.second_text.text().strip()}

        text = self.first_text.text().strip()
        return text and os.path.exists(text) and os.path.isfile(text)

    def get_value(self, name=""):
        mask_type = self.select_type.currentEnum()
        if mask_type == FileMaskType.Suffix:
            return MaskSuffix(name=name, suffix=self.first_text.text().strip())
        if mask_type == FileMaskType.Replace:
            return MaskSub(name=name, base=self.first_text.text().strip(), rep=self.second_text.text().strip())
        return MaskFile(name=name, path_to_file=self.first_text.text().strip())


class MaskOperation(Enum):
    mask_intersection = 0
    mask_sum = 1

    def __str__(self):
        return self.name.replace("_", " ").capitalize()


class ProtectedGroupBox(QGroupBox):
    def __init__(self, text: str, parent: typing.Optional[QWidget] = None):
        super().__init__(text, parent)
        self.setStyleSheet(group_sheet)
        self.protect = False
        self._node_type = None
        self._parent_node_type = None
        self._replace = False

    def set_current_node(self, node: typing.Optional[NodeType], parent_node: typing.Optional[NodeType] = None):
        self._node_type = node
        self._parent_node_type = parent_node
        self._activate_button()

    def set_replace(self, replace: bool):
        self._replace = replace
        self._activate_button()

    def _activate_button(self, _value=None):
        raise NotImplementedError

    @contextmanager
    def enable_protect(self):
        previous = self.protect
        self.protect = True
        try:
            yield
        finally:
            self.protect = previous

    @classmethod
    def refresh_profiles(
        cls, list_widget: typing.Union[QListWidget, SearchableListWidget], new_values: typing.List[str]
    ):
        index = cls.get_index(list_widget.currentItem(), new_values)
        list_widget.clear()
        list_widget.addItems(new_values)
        if index != -1:
            list_widget.setCurrentRow(index)

    @staticmethod
    def get_index(item: QListWidgetItem, new_values: typing.List[str]) -> int:
        if item is None:
            return -1
        text = item.text()
        try:
            return new_values.index(text)
        except ValueError:  # pragma: no cover
            return -1


class OtherOperations(ProtectedGroupBox):
    save_operation = Signal(object)

    def __init__(self, parent=None):
        super().__init__("Other operations:", parent)
        self.save_translate_dict: typing.Dict[str, SaveBase] = {x.get_short_name(): x for x in save_dict.values()}
        self.save_constructor = None

        self.change_root = QEnumComboBox(self, enum_class=RootType)
        self.choose_save_method = QComboBox()
        self.choose_save_method.addItem("<none>")
        self.choose_save_method.addItems(list(self.save_translate_dict.keys()))
        self.save_btn = QPushButton("Save")

        self.choose_save_method.currentTextChanged.connect(self.save_changed)
        self.save_btn.clicked.connect(self.save_action)

        layout = QVBoxLayout()
        layout.setSpacing(0)
        layout.addWidget(QLabel("Root type:"))
        layout.addWidget(self.change_root)
        layout.addStretch(1)
        layout.addWidget(QLabel("Saving:"))
        layout.addWidget(self.choose_save_method)
        layout.addWidget(self.save_btn)

        self.setLayout(layout)

    @property
    def root_type_changed(self):
        return self.change_root.currentEnumChanged

    def save_changed(self, text):
        text = str(text)
        save_class = self.save_translate_dict.get(text, None)
        if save_class is None:
            self.choose_save_method.setCurrentText("<none>")
            self.save_btn.setText("Save")
            self.save_btn.setToolTip("Choose file type")
        else:
            self.save_btn.setText(f"Save to {save_class.get_short_name()}")
            self.save_btn.setToolTip("Choose mask create in plan view")
        self._activate_button()

    @property
    def expected_node_type(self) -> typing.Optional[NodeType]:
        save_class = self.save_translate_dict.get(self.choose_save_method.currentText(), None)
        if save_class is None:
            return None
        if save_class.need_mask():
            return NodeType.mask
        return NodeType.segment if save_class.need_segmentation() else NodeType.root

    def _activate_button(self, _value=None):
        if self._replace:
            self.save_btn.setEnabled(self._parent_node_type == self.expected_node_type and self._node_type is not None)
        else:
            self.save_btn.setEnabled(self._node_type == self.expected_node_type and self._node_type is not None)

    def save_action(self):
        save_class = self.save_translate_dict.get(self.choose_save_method.currentText(), None)
        if save_class is None:  # pragma: no cover
            show_warning(self, "Save problem", "Not found save class")
            return
        dial = FormDialog(
            [
                AlgorithmProperty("suffix", "File suffix", ""),
                AlgorithmProperty("directory", "Sub directory", ""),
                *save_class.get_fields(),
            ]
        )
        if not dial.exec_():
            return
        values = dial.get_values()
        suffix = values["suffix"]
        directory = values["directory"]
        del values["suffix"]
        del values["directory"]
        save_elem = Save(
            suffix=suffix,
            directory=directory,
            algorithm=save_class.get_name(),
            short_name=save_class.get_short_name(),
            values=values,
        )
        self.save_operation.emit(save_elem)


class ROIExtractionOp(ProtectedGroupBox):
    roi_extraction_profile_selected = Signal(object)
    roi_extraction_pipeline_selected = Signal(object)
    roi_extraction_profile_add = Signal(object)
    roi_extraction_pipeline_add = Signal(object)

    def __init__(self, settings: PartSettings, parent: typing.Optional[QWidget] = None):
        super().__init__("ROI extraction", parent)
        self.settings = settings

        self.roi_profile = SearchableListWidget()
        self.roi_pipeline = SearchableListWidget()
        self.roi_extraction_tab = QTabWidget()
        self.roi_extraction_tab.addTab(self.roi_profile, "Profile")
        self.roi_extraction_tab.addTab(self.roi_pipeline, "Pipeline")

        self.choose_profile_btn = QPushButton("Add Profile")
        self.choose_profile_btn.setDisabled(True)

        self.settings.roi_profiles_changed.connect(self._refresh_profiles)
        self.settings.roi_pipelines_changed.connect(self._refresh_pipelines)

        self.roi_profile.currentTextChanged.connect(self._roi_extraction_profile_selected)
        self.roi_pipeline.currentTextChanged.connect(self._roi_extraction_pipeline_selected)
        self.choose_profile_btn.clicked.connect(self._add_profile)
        self.roi_extraction_tab.currentChanged.connect(self._on_change_tab)

        layout = QVBoxLayout()
        layout.setSpacing(0)
        layout.addWidget(self.roi_extraction_tab)
        layout.addWidget(self.choose_profile_btn)

        self.setLayout(layout)

        self._refresh_profiles()
        self._refresh_pipelines()
        self._update_btn_text()
        self.settings.roi_profiles_changed.connect(self._refresh_profiles)
        self.settings.roi_pipelines_changed.connect(self._refresh_pipelines)

    def set_replace(self, replace: bool):
        super().set_replace(replace)
        self._on_change_tab()

    def _activate_button(self, _value=None):
        if self._replace:
            self.choose_profile_btn.setEnabled(
                self._node_type == NodeType.segment
                and self.roi_extraction_tab.currentWidget() == self.roi_profile
                and self.roi_profile.currentItem() is not None
            )
            return
        self.choose_profile_btn.setEnabled(
            self._node_type in {NodeType.root, NodeType.mask, NodeType.file_mask}
            and self.roi_extraction_tab.currentWidget().currentRow() >= 0
        )

    def _update_btn_text(self):
        index = self.roi_extraction_tab.currentIndex()
        text = self.roi_extraction_tab.tabText(index)
        if self._replace:
            self.choose_profile_btn.setText(f"Replace {text}")
        else:
            self.choose_profile_btn.setText(f"Add {text}")

    def _on_change_tab(self, _val=None):
        self._update_btn_text()
        with self.enable_protect():
            self.roi_profile.setCurrentItem(None)
            self.roi_pipeline.setCurrentItem(None)
        self._activate_button()

    def _refresh_profiles(self):
        new_profiles = sorted(self.settings.roi_profiles.keys(), key=str.lower)
        with self.enable_protect():
            self.refresh_profiles(self.roi_profile, new_profiles)

    def _refresh_pipelines(self):
        new_pipelines = sorted(self.settings.roi_pipelines.keys(), key=str.lower)
        with self.enable_protect():
            self.refresh_profiles(self.roi_pipeline, new_pipelines)

    def _roi_extraction_profile_selected(self, name: str):
        if self.protect:
            return
        self._activate_button()
        self.roi_extraction_profile_selected.emit(self.settings.roi_profiles[name])

    def _roi_extraction_pipeline_selected(self, name: str):
        if self.protect:
            return
        self._activate_button()
        self.roi_extraction_pipeline_selected.emit(self.settings.roi_pipelines[name])

    def _add_profile(self):
        if self.roi_extraction_tab.currentWidget() == self.roi_profile:
            item = self.roi_profile.currentItem()
            if item is None:
                return
            self.roi_extraction_profile_add.emit(deepcopy(self.settings.roi_profiles[item.text()]))
        else:
            item = self.roi_pipeline.currentItem()
            if item is None:
                return
            self.roi_extraction_pipeline_add.emit(deepcopy(self.settings.roi_pipelines[item.text()]))


class SelectMeasurementOp(ProtectedGroupBox):
    set_of_measurement_add = Signal(object)
    set_of_measurement_selected = Signal(object)

    def __init__(self, settings: PartSettings, parent: typing.Optional[QWidget] = None):
        super().__init__("Set of measurements:", parent)
        self.settings = settings

        self.measurements_list = SearchableListWidget(self)
        self.measurement_name_prefix = QLineEdit(self)
        self.choose_channel_for_measurements = QComboBox()
        self.choose_channel_for_measurements.addItems(
            ["Same as segmentation"] + [str(x + 1) for x in range(MAX_CHANNEL_NUM)]
        )
        self.units_choose = QEnumComboBox(enum_class=Units)
        self.units_choose.setCurrentEnum(self.settings.get("units_value", Units.nm))
        self.add_measurement_btn = QPushButton("Add measurement calculation")
        self.add_measurement_btn.clicked.connect(self._measurement_add)
        self.measurements_list.currentTextChanged.connect(self._measurement_selected)

        layout = QGridLayout()
        layout.setSpacing(0)
        layout.addWidget(self.measurements_list, 0, 0, 1, 2)
        lab = QLabel("Name prefix:")
        lab.setToolTip("Prefix added before each column name")
        layout.addWidget(lab, 1, 0)
        layout.addWidget(self.measurement_name_prefix, 1, 1)
        layout.addWidget(QLabel("Channel:"), 2, 0)
        layout.addWidget(self.choose_channel_for_measurements, 2, 1)
        layout.addWidget(QLabel("Units:"), 3, 0)
        layout.addWidget(self.units_choose, 3, 1)
        layout.addWidget(self.add_measurement_btn, 4, 0, 1, 2)
        self.setLayout(layout)

        self.add_measurement_btn.setDisabled(True)
        self._refresh_measurement()
        self.settings.measurement_profiles_changed.connect(self._refresh_measurement)

    def set_replace(self, replace: bool):
        super().set_replace(replace)
        self.add_measurement_btn.setText("Replace set of measurements" if self._replace else "Add set of measurements")

    def _activate_button(self, _value=None):
        if self._replace:
            self.add_measurement_btn.setEnabled(
                self._node_type == NodeType.measurement and self.measurements_list.currentItem() is not None
            )
        else:
            self.add_measurement_btn.setEnabled(
                self._node_type == NodeType.segment and self.measurements_list.currentItem() is not None
            )

    def _refresh_measurement(self):
        new_measurements = sorted(self.settings.measurement_profiles.keys(), key=str.lower)
        with self.enable_protect():
            self.refresh_profiles(self.measurements_list, new_measurements)

    def _measurement_add(self):
        item = self.measurements_list.currentItem()
        if item is None:
            return
        measurement_copy = deepcopy(self.settings.measurement_profiles[item.text()])
        prefix = str(self.measurement_name_prefix.text()).strip()
        channel = self.choose_channel_for_measurements.currentIndex() - 1
        measurement_copy.name_prefix = prefix
        self.set_of_measurement_add.emit(
            MeasurementCalculate(
                channel=channel,
                measurement_profile=measurement_copy,
                name_prefix=prefix,
                units=self.units_choose.currentEnum(),
            )
        )

    def _measurement_selected(self, name: str):
        if self.protect:
            return
        self._activate_button()
        self.set_of_measurement_selected.emit(self.settings.measurement_profiles[name])


class StretchWrap(QWidget):
    def __init__(self, widget: QWidget, parent: typing.Optional[QWidget] = None):
        super().__init__(parent)
        self.widget = widget
        lay = QVBoxLayout()
        lay.setSpacing(0)
        lay.addWidget(widget)
        lay.addStretch(1)
        self.setLayout(lay)

    def __getattr__(self, item):
        return getattr(self.widget, item)


class SelectMaskOp(ProtectedGroupBox):
    mask_step_add = Signal(object)

    def __init__(self, settings: PartSettings, parent: typing.Optional[QWidget] = None):
        super().__init__("Use mask from:", parent)
        self.settings = settings
        self.mask_set = {}

        self.file_mask = FileMask()
        self.mask_from_segmentation = MaskWidget(settings)
        self.mask_operation = StretchWrap(QEnumComboBox(enum_class=MaskOperation))
        self.add_mask_btn = QPushButton("Add mask")
        self.add_mask_btn.setToolTip("Mask need to have unique name")
        self.add_mask_btn.clicked.connect(self._add_mask)
        self.mask_name = QLineEdit()

        self.mask_tab_select = QTabWidget()

        self.mask_tab_select.addTab(self.file_mask, "File")
        self.mask_tab_select.addTab(self.mask_from_segmentation, "Current ROI")
        self.mask_tab_select.addTab(self.mask_operation, "Operations on masks")
        self.mask_tab_select.setTabToolTip(2, "Allows to create mask which is based on masks previously added to plan.")
        self.mask_tab_select.currentChanged.connect(self._activate_button)
        self.mask_name.textChanged.connect(self._activate_button)

        layout = QGridLayout()
        layout.setSpacing(0)
        layout.addWidget(self.mask_tab_select, 0, 0, 1, 2)
        label = QLabel("Mask name:")
        label.setToolTip("Needed if you would like to reuse this mask in tab 'Operations on masks'")
        self.mask_name.setToolTip("Needed if you would like to reuse this mask in tab 'Operations on masks'")
        layout.addWidget(label, 1, 0)
        layout.addWidget(self.mask_name, 1, 1)
        layout.addWidget(self.add_mask_btn, 2, 0, 1, 2)
        self.setLayout(layout)

        self.add_mask_btn.setDisabled(True)

    def update_mask_set(self, mask_set: typing.Set[str]):
        self.mask_set = mask_set

    def set_replace(self, replace: bool):
        super().set_replace(replace)
        self.add_mask_btn.setText("Replace mask" if self._replace else "Add mask")

    def _activate_button(self, _value=None):
        name = self.mask_name.text().strip()
        name_ok = not name or name not in self.mask_set
        if self._replace:
            name_ok = name_ok and self._node_type == NodeType.mask
            node_type = self._parent_node_type
        else:
            node_type = self._node_type
        if self.mask_tab_select.currentWidget() == self.mask_from_segmentation:
            self.add_mask_btn.setEnabled(node_type == NodeType.segment and name_ok)
            return
        self.add_mask_btn.setEnabled(node_type == NodeType.root and name_ok)

    def _add_mask(self):
        widget = self.mask_tab_select.currentWidget()
        name = self.mask_name.text().strip()
        if widget == self.file_mask:
            mask_ob = self.file_mask.get_value(name)
        elif widget == self.mask_from_segmentation:
            mask_ob = MaskCreate(name=name, mask_property=self.mask_from_segmentation.get_mask_property())
        elif widget == self.mask_operation:
            dial = TwoMaskDialog(self.mask_set)
            if not dial.exec_():
                return  # pragma: no cover
            names = dial.get_result()

            if self.mask_operation.currentEnum() == MaskOperation.mask_intersection:  # Mask intersection
                mask_construct = MaskIntersection
            else:
                mask_construct = MaskSum
            mask_ob = mask_construct(name=name, mask1=names[0], mask2=names[1])
        else:
            raise ValueError("Unknown widget")  # pragma: no cover

        self.mask_step_add.emit(mask_ob)


class CreatePlan(QWidget):
    plan_node_changed = Signal()

    def __init__(self, settings: PartSettings):
        super().__init__()
        self.settings = settings
        self.save_translate_dict: typing.Dict[str, SaveBase] = {x.get_short_name(): x for x in save_dict.values()}
        self._mask_set = set()
        self.plan = PlanPreview(self)
        self.save_plan_btn = QPushButton("Save")
        self.clean_plan_btn = QPushButton("Remove all")
        self.remove_btn = QPushButton("Remove")
        self.update_element_chk = QCheckBox("Update element")
        self.other_operations = OtherOperations(self)
        self.roi_extraction = ROIExtractionOp(settings=settings, parent=self)
        self.select_measurement = SelectMeasurementOp(settings=settings, parent=self)
        self.select_mask = SelectMaskOp(settings=settings, parent=self)
        self.mask_set = set()

        self.expected_node_type = None
        self.save_constructor = None

        self.information = QTextEdit()
        self.information.setReadOnly(True)

        # FIXME: fix in better way
        self.calculation_plan = CalculationPlan()
        self.plan.set_plan(self.calculation_plan)
        self.segmentation_mask = MaskWidget(settings)
        self.file_mask = FileMask()

        self.other_operations.root_type_changed.connect(self.change_root_type)
        self.other_operations.save_operation.connect(self.add_save_operation)
        self.roi_extraction.roi_extraction_pipeline_selected.connect(self.show_info)
        self.roi_extraction.roi_extraction_profile_selected.connect(self.show_info)
        self.roi_extraction.roi_extraction_profile_add.connect(self.add_roi_extraction)
        self.roi_extraction.roi_extraction_pipeline_add.connect(self.add_roi_extraction_pipeline)
        self.select_measurement.set_of_measurement_add.connect(self.add_set_of_measurement)
        self.select_measurement.set_of_measurement_selected.connect(self.show_info)
        self.select_mask.mask_step_add.connect(self.create_mask)

        self.clean_plan_btn.clicked.connect(self.clean_plan)
        self.remove_btn.clicked.connect(self.remove_element)
        self.save_plan_btn.clicked.connect(self.add_calculation_plan)
        self.update_element_chk.stateChanged.connect(self.select_mask.set_replace)
        self.update_element_chk.stateChanged.connect(self.roi_extraction.set_replace)
        self.update_element_chk.stateChanged.connect(self.select_measurement.set_replace)

        self.setup_ui()

        self.node_type = NodeType.root
        self.node_name = ""
        self.plan.changed_node.connect(self.node_type_changed)
        self.node_type_changed()

    def setup_ui(self):
        plan_box = QGroupBox("Prepare workflow:")
        lay = QVBoxLayout()
        lay.addWidget(self.plan)
        bt_lay = QGridLayout()
        bt_lay.setSpacing(1)
        bt_lay.addWidget(self.save_plan_btn, 0, 0)
        bt_lay.addWidget(self.clean_plan_btn, 0, 1)
        bt_lay.addWidget(self.remove_btn, 1, 0)
        bt_lay.addWidget(self.update_element_chk, 1, 1)
        lay.addLayout(bt_lay)
        plan_box.setLayout(lay)
        plan_box.setStyleSheet(group_sheet)

        info_box = QGroupBox("Information")
        info_box.setStyleSheet(group_sheet)
        lay = QVBoxLayout()
        lay.addWidget(self.information)
        info_box.setLayout(lay)

        layout = QGridLayout()
        layout.addWidget(plan_box, 0, 0, 5, 1)
        layout.addWidget(self.select_mask, 0, 2, 1, 2)
        layout.addWidget(self.other_operations, 0, 1)
        layout.addWidget(self.roi_extraction, 1, 1, 1, 2)
        layout.addWidget(self.select_measurement, 1, 3)
        layout.addWidget(info_box, 3, 1, 1, 3)
        self.setLayout(layout)

    @property
    def mask_set(self):
        return self._mask_set

    @mask_set.setter
    def mask_set(self, value):
        self._mask_set = value
        self.select_mask.update_mask_set(value)

    def change_root_type(self, root_type: RootType):
        self.calculation_plan.set_root_type(root_type)
        self.plan.update_view()

    def add_save_operation(self, save_info: Save):
        if self.update_element_chk.isChecked():
            self.calculation_plan.replace_step(save_info)
        else:
            self.calculation_plan.add_step(save_info)
        self.plan.update_view()

    def add_set_of_measurement(self, set_of_measurement: MeasurementCalculate):
        if self.update_element_chk.isChecked():
            self.calculation_plan.replace_step(set_of_measurement)
        else:
            self.calculation_plan.add_step(set_of_measurement)
        self.plan.update_view()

    def node_type_changed(self):
        self.node_name = ""
        if self.plan.currentItem() is None:
            self.remove_btn.setDisabled(True)
            self.plan_node_changed.emit()
            logging.debug("[node_type_changed] return")
            return
        node_type = self.calculation_plan.get_node_type()

        node_type_for_ob = self.calculation_plan.get_node_type(parent=self.update_element_chk.isChecked())

        self.other_operations.set_current_node(node_type, node_type_for_ob)
        self.roi_extraction.set_current_node(node_type, node_type_for_ob)
        self.select_measurement.set_current_node(node_type, node_type_for_ob)
        self.select_mask.set_current_node(node_type, node_type_for_ob)

        self.node_type = node_type
        self.plan_node_changed.emit()

    def create_mask(self, mask_ob: MaskBase):
        if mask_ob.name and mask_ob.name in self.mask_set:
            show_warning("Already exists", "Mask with this name already exists")
            return

        if self.update_element_chk.isChecked():
            node = self.calculation_plan.get_node()
            name = node.operation.name
            if name in self.calculation_plan.get_reused_mask() and name != mask_ob.name:
                show_warning(
                    "Cannot remove", f"Cannot remove mask '{name}' from plan because it is used in other elements"
                )
                return

            self.mask_set.remove(name)
            self.mask_set.add(mask_ob.name)
            self.calculation_plan.replace_step(mask_ob)
        else:
            self.mask_set.add(mask_ob.name)
            self.calculation_plan.add_step(mask_ob)
        self.plan.update_view()

    def add_roi_extraction(self, roi_extraction: ROIExtractionOp):
        if self.update_element_chk.isChecked():
            self.calculation_plan.replace_step(roi_extraction)
        else:
            self.calculation_plan.add_step(roi_extraction)
        self.plan.update_view()

    def add_roi_extraction_pipeline(self, roi_extraction_pipeline: SegmentationPipeline):
        if self.update_element_chk.isChecked():
            show_warning("Cannot update pipeline", "Cannot update pipeline")
            return
        pos = self.calculation_plan.current_pos[:]
        old_pos = pos[:]
        for el in roi_extraction_pipeline.mask_history:
            self.calculation_plan.add_step(el.segmentation)
            self.plan.update_view()
            node = self.calculation_plan.get_node(pos)
            pos.append(len(node.children) - 1)
            self.calculation_plan.set_position(pos)
            self.calculation_plan.add_step(MaskCreate(name="", mask_property=el.mask_property))
            self.plan.update_view()
            pos.append(0)
            self.calculation_plan.set_position(pos)
        self.calculation_plan.add_step(roi_extraction_pipeline.segmentation)
        self.calculation_plan.set_position(old_pos)
        self.plan.update_view()

    def remove_element(self):
        conflict_mask, used_mask = self.calculation_plan.get_file_mask_names()
        if len(conflict_mask) > 0:
            logging.info("Mask in use")
            show_warning("In use", f'Masks {", ".join(conflict_mask)} are used in other places')

            return
        self.mask_set -= used_mask
        self.calculation_plan.remove_step()
        self.plan.update_view()

    def clean_plan(self):
        self.calculation_plan = CalculationPlan()
        self.plan.set_plan(self.calculation_plan)
        self.node_type_changed()
        self.mask_set = set()

    def add_calculation_plan(self, text=None):
        if text is None or isinstance(text, bool):
            text, ok = QInputDialog.getText(self, "Plan title", "Set plan title")
        else:
            text, ok = QInputDialog.getText(
                self, "Plan title", f"Set plan title. Previous ({text}) is already in use", text=text
            )
        text = text.strip()
        if ok:
            if not text:
                QMessageBox.information(
                    self, "Name cannot be empty", "Name cannot be empty, Please set correct name", QMessageBox.Ok
                )
                self.add_calculation_plan()
                return
            if text in self.settings.batch_plans:
                res = QMessageBox.information(
                    self,
                    "Name already in use",
                    "Name already in use. Would like to overwrite?",
                    QMessageBox.Yes | QMessageBox.No,
                )
                if res == QMessageBox.No:
                    self.add_calculation_plan(text)
                    return
            plan = copy(self.calculation_plan)
            plan.set_name(text)
            self.settings.batch_plans[text] = plan
            self.settings.dump()

    def show_info(self, item: typing.Union[ROIExtractionOp, SegmentationPipeline, MeasurementProfile]):
        if isinstance(item, (ROIExtractionOp, MeasurementProfile)):
            self.information.setText(str(item))
        else:
            self.information.setText(item.pretty_print(AnalysisAlgorithmSelection))

    def edit_plan(self, plan: CalculationPlan):
        if plan.is_bad():
            QMessageBox().warning(
                self, "Cannot edit broken plan", f"Cannot edit broken plan. {plan.get_error_source()}"
            )
            return
        self.calculation_plan = copy(plan)
        self.plan.set_plan(self.calculation_plan)
        self.mask_set.clear()
        self.calculation_plan.set_position([])
        self.mask_set.update(self.calculation_plan.get_mask_names())


class PlanPreview(QTreeWidget):
    """
    :type calculation_plan: CalculationPlan
    """

    changed_node = Signal()

    def __init__(self, parent=None, calculation_plan=None):
        super().__init__(parent)
        self.calculation_plan = None
        self.header().close()
        self.itemSelectionChanged.connect(self.set_path)
        self.setContextMenuPolicy(Qt.ContextMenuPolicy.CustomContextMenu)
        if calculation_plan is not None:
            self.set_plan(calculation_plan)

    def restore_path(self, widget, path):
        """
        :type widget: QTreeWidgetItem
        :type path: list[int]
        :param widget:
        :param path:
        :return:
        """
        if widget is None:
            return list(reversed(path))
        parent = widget.parent()
        if parent is None:
            return list(reversed(path))
        index = parent.indexOfChild(widget)
        if str(parent.child(0).text(0)) == "Description":
            index -= 1
        if index == -1:
            return None
        path.append(index)
        return self.restore_path(parent, path)

    def set_path(self):
        current_item = self.currentItem()  # type : QTreeWidgetItem
        if current_item is None:
            return
        self.calculation_plan.set_position(self.restore_path(current_item, []))
        self.changed_node.emit()

    def preview_object(self, calculation_plan):
        self.set_plan(calculation_plan)

    def set_plan(self, calculation_plan):
        if calculation_plan is not None and calculation_plan.is_bad():
            QMessageBox().warning(
                self, "Cannot preview broken plan", f"Cannot preview broken plan. {calculation_plan.get_error_source()}"
            )
            return
        self.calculation_plan = calculation_plan
        self.setCurrentItem(self.topLevelItem(0))
        self.update_view(True)

    def explore_tree(self, up_widget, node_plan, deep=True):
        """
        :type up_widget: QTreeWidgetItem
        :type node_plan: CalculationTree
        :type deep: bool
        :param up_widget: List widget item
        :param node_plan: node from calculation plan
        :return:
        """
        widget = QTreeWidgetItem(up_widget)
        widget.setText(0, CalculationPlan.get_el_name(node_plan.operation))
        self.setCurrentItem(widget)
        if isinstance(node_plan.operation, (MeasurementCalculate, ROIExtractionProfile)):
            widget.setData(0, Qt.ItemDataRole.UserRole, node_plan.operation)
        if isinstance(node_plan.operation, (MeasurementCalculate, ROIExtractionProfile, MaskCreate)):
            desc = QTreeWidgetItem(widget)
            desc.setText(0, "Description")
            if isinstance(node_plan.operation, ROIExtractionProfile):
                txt = node_plan.operation.pretty_print(AnalysisAlgorithmSelection)
            else:
                txt = str(node_plan.operation)
            for line in txt.split("\n")[1:]:
                QTreeWidgetItem(desc, [line])
        if deep:
            for el in node_plan.children:
                self.explore_tree(widget, el)
        up_widget.setExpanded(True)

    def get_node(self, path):
        """
        :type path: list[int]
        :param path:
        :return: QTreeWidgetItem
        """
        widget = self.topLevelItem(0)  # type : QTreeWidgetItem
        for index in path:
            if str(widget.child(0).text(0)) == "Description":
                widget = widget.child(index + 1)
            else:
                widget = widget.child(index)
        return widget

    def update_view(self, reset=False):
        if reset:
            self.clear()
            root = QTreeWidgetItem(self)
            root.setText(0, f"Root {self.calculation_plan.get_root_type()}")
            self.setCurrentItem(root)
            for el in self.calculation_plan.execution_tree.children:
                self.explore_tree(root, el, True)
            return
        self.blockSignals(True)
        root = self.get_node([])
        root.setText(0, f"Root {self.calculation_plan.get_root_type()}")
        for path, el, op_type in self.calculation_plan.get_changes():
            if op_type == PlanChanges.add_node:
                node = self.get_node(path)
                self.explore_tree(node, el, False)
            elif op_type == PlanChanges.remove_node:
                node = self.get_node(path[:-1])
                index = path[-1]
                if str(node.child(0).text(0)) == "Description":
                    index += 1
                node.removeChild(node.child(index))
            elif op_type == PlanChanges.replace_node:
                node = self.get_node(path)
                node.setText(0, CalculationPlan.get_el_name(el.operation))
                if isinstance(el.operation, (MeasurementCalculate, ROIExtractionProfile, MaskCreate)):
                    child = node.child(0)
                    child.takeChildren()
                    if isinstance(el.operation, ROIExtractionProfile):
                        txt = el.operation.pretty_print(AnalysisAlgorithmSelection)
                    else:
                        txt = str(el.operation)
                    for line in txt.split("\n")[1:]:
                        QTreeWidgetItem(child, [line])

            else:
                logging.error("Unknown operation %s", op_type)  # pragma: no cover
        self.blockSignals(False)
        self.set_path()
        self.changed_node.emit()


class CalculateInfo(QWidget):
    """
    "widget to show information about plans and allow to se plan details
    :type settings: Settings
    """

    plan_to_edit_signal = Signal(object)

    def __init__(self, settings: PartSettings):
        super().__init__()
        self.settings = settings
        self.calculate_plans = SearchableListWidget(self)
        self.plan_view = PlanPreview(self)
        self.delete_plan_btn = QPushButton("Delete")
        self.edit_plan_btn = QPushButton("Edit")
        self.export_plans_btn = QPushButton("Export")
        self.import_plans_btn = QPushButton("Import")
        info_layout = QVBoxLayout()
        info_butt_layout = QGridLayout()
        info_butt_layout.setSpacing(1)
        info_butt_layout.addWidget(self.delete_plan_btn, 1, 1)
        info_butt_layout.addWidget(self.edit_plan_btn, 0, 1)
        info_butt_layout.addWidget(self.export_plans_btn, 1, 0)
        info_butt_layout.addWidget(self.import_plans_btn, 0, 0)
        info_layout.addLayout(info_butt_layout)
        info_chose_layout = QVBoxLayout()
        info_chose_layout.setSpacing(2)
        info_chose_layout.addWidget(QLabel("List of workflows:"))
        info_chose_layout.addWidget(self.calculate_plans)
        info_chose_layout.addWidget(QLabel("Preview:"))
        info_chose_layout.addWidget(self.plan_view)
        info_layout.addLayout(info_chose_layout)
        self.setLayout(info_layout)
        self.protect = False
        self.plan_to_edit = None

        self.plan_view.header().close()
        self.calculate_plans.currentTextChanged.connect(self.plan_preview)
        self.delete_plan_btn.clicked.connect(self.delete_plan)
        self.edit_plan_btn.clicked.connect(self.edit_plan)
        self.export_plans_btn.clicked.connect(self.export_plans)
        self.import_plans_btn.clicked.connect(self.import_plans)
        self.settings.batch_plans_changed.connect(self.update_plan_list)
        self.plan_view.customContextMenuRequested.connect(self._context_menu)
        self.update_plan_list()

    def _context_menu(self, point):
        item = self.plan_view.itemAt(point)
        data = item.data(0, Qt.ItemDataRole.UserRole)
        if data is None:
            return

        menu = QMenu(self)
        if isinstance(data, ROIExtractionProfile):
            action = QAction("Save ROI extraction Profile")
            action.triggered.connect(lambda _: self._save_roi_profile(data))
        elif isinstance(data, MeasurementCalculate):
            action = QAction("Save Measurement profile")
            action.triggered.connect(lambda _: self._save_measurement_profile(data.measurement_profile))
        else:
            raise ValueError(f"Not supported data type {type(data)} for {data}")
        menu.addAction(action)
        menu.exec_(self.plan_view.mapToGlobal(point))

    def _save_roi_profile(self, data: ROIExtractionProfile):
        if data.name in self.settings.roi_profiles:
            text, ok = QInputDialog.getText(
                self, "Name collision", "Profile with this name exists, please provide a new name.", text=data.name
            )
            if not ok:
                return None
            return self._save_roi_profile(typing.cast(ROIExtractionProfile, data.copy(update={"name": text})))
        self.settings.roi_profiles[data.name] = data
        return None

    def _save_measurement_profile(self, data: MeasurementProfile):
        if data.name in self.settings.measurement_profiles:
            text, ok = QInputDialog.getText(
                self, "Name collision", "Profile with this name exists, please provide a new name.", text=data.name
            )
            if not ok:
                return None
            return self._save_measurement_profile(typing.cast(MeasurementProfile, data.copy(update={"name": text})))
        self.settings.measurement_profiles[data.name] = data
        return None

    def update_plan_list(self):
        new_plan_list = sorted(self.settings.batch_plans.items(), key=lambda x: x[0])
        if self.calculate_plans.currentItem() is not None:
            text = str(self.calculate_plans.currentItem().text())
            try:
                index = [x[0] for x in new_plan_list].index(text)
            except ValueError:
                index = -1
        else:
            index = -1
        self.protect = True
        self.calculate_plans.clear()

        for name, plan in new_plan_list:
            item = QListWidgetItem(name)
            if plan.is_bad():
                item.setIcon(QIcon(os.path.join(icons_dir, "task-reject.png")))
                item.setToolTip(plan.get_error_source())
            self.calculate_plans.addItem(item)
        if index != -1:
            self.calculate_plans.setCurrentRow(index)
        self.protect = False

    def export_plans(self):
        choose = ExportDialog(self.settings.batch_plans, PlanPreview)
        if not choose.exec_():
            return
        dial = PSaveDialog(
            "Calculation plans (*.json)",
            caption="Export calculation plans",
            settings=self.settings,
            path="io.batch_plan_directory",
        )
        dial.selectFile("calculation_plans.json")
        if dial.exec_():
            file_path = str(dial.selectedFiles()[0])
            data = {x: self.settings.batch_plans[x] for x in choose.get_export_list()}
            with open(file_path, "w", encoding="utf-8") as ff:
                json.dump(data, ff, cls=self.settings.json_encoder_class, indent=2)

    def import_plans(self):
        dial = PLoadDialog(
            [LoadPlanJson, LoadPlanExcel],
            settings=self.settings,
            path="io.batch_plan_directory",
            caption="Import calculation plans",
        )
        if dial.exec_():
            res = dial.get_result()
            plans, err = res.load_class.load(res.load_location)
            if err:
                error_str = "\n".join(err)
                show_warning("Import error", f"error during importing, part of data were filtered. {error_str}")
            if not plans:
                show_warning("Import error", "No plans were imported")
                return
            choose = ImportDialog(plans, self.settings.batch_plans, PlanPreview, CalculationPlan)
            if choose.exec_():
                for original_name, final_name in choose.get_import_list():
                    self.settings.batch_plans[final_name] = plans[original_name]

    def delete_plan(self):
        if self.calculate_plans.currentItem() is None:
            return
        text = str(self.calculate_plans.currentItem().text())
        if not text:
            return  # pragma: no cover
        if text in self.settings.batch_plans:
            del self.settings.batch_plans[text]
        self.plan_view.clear()

    def edit_plan(self):
        if self.calculate_plans.currentItem() is None:
            return
        text = str(self.calculate_plans.currentItem().text())
        if not text:
            return  # pragma: no cover
        if text in self.settings.batch_plans:
            self.plan_to_edit = self.settings.batch_plans[text]
            self.plan_to_edit_signal.emit(self.plan_to_edit)

    def plan_preview(self, text):
        if self.protect:
            return
        text = str(text)
        if not text.strip():
            return
        plan = self.settings.batch_plans[text]
        if not plan.is_bad():
            self.plan_view.set_plan(plan)


class CalculatePlaner(QSplitter):
    """
    :type settings: Settings
    """

    def __init__(self, settings, parent=None):
        super().__init__(parent)
        self.settings = settings
        self.info_widget = CalculateInfo(settings)
        self.addWidget(self.info_widget)
        self.create_plan = CreatePlan(settings)
        self.info_widget.plan_to_edit_signal.connect(self.create_plan.edit_plan)
        self.addWidget(self.create_plan)



import typing
from textwrap import indent

from PartSegCore.algorithm_describe_base import ROIExtractionProfile
from PartSegCore.mask_create import MaskProperty
from PartSegCore.utils import BaseModel


class SegmentationPipelineElement(BaseModel):
    segmentation: ROIExtractionProfile
    mask_property: MaskProperty

    def pretty_print(self, algorithm_dict):
        return (
            indent("Segmentation:\n" + self.segmentation.pretty_print(algorithm_dict).split("\n", 1)[1], "    ")
            + "\n\n"
            + indent(str(self.mask_property), "    ")
        )

    def __str__(self):
        return indent(str(self.segmentation), "    ") + "\n\n" + indent(str(self.mask_property), "    ")

    def __repr__(self):
        return f"SegmentationPipelineElement(segmentation={self.segmentation!r},mask_property={self.mask_property!r})"


class SegmentationPipeline(BaseModel):
    name: str
    segmentation: ROIExtractionProfile
    mask_history: typing.List[SegmentationPipelineElement]

    def pretty_print(self, algorithm_dict):
        return (
            (
                f"Segmentation pipeline name: {self.name}\n"
                + "\n--------------\n".join(x.pretty_print(algorithm_dict) for x in self.mask_history)
            )
            + "\n--------------\nLast segmentation:\n"
        ) + self.segmentation.pretty_print(algorithm_dict).split("\n", 1)[1]

    def __str__(self):
        return (
            (
                f"Segmentation pipeline name: {self.name}\n"
                + "\n--------------\n".join(str(x) for x in self.mask_history)
            )
            + "\n--------------\nLast segmentation\n"
        ) + str(self.segmentation)

    def __repr__(self):
        return (
            f"SegmentationPipeline(name={self.name},\nmask_history={self.mask_history},\n"
            f"segmentation={self.segmentation!r})"
        )



import qtawesome as qta
from qtpy.QtGui import QColor
from qtpy.QtWidgets import QPushButton

from PartSeg.common_backend.base_settings import BaseSettings


class QtViewerPushButton(QPushButton):
    def __init__(self, icon: str, settings: BaseSettings):
        super().__init__(qta.icon(icon), "")
        self.settings = settings
        settings.theme_changed.connect(self._theme_changed)
        self._theme_changed()

    def _theme_changed(self):
        color = self.settings.theme.text
        self.setIcon(qta.icon("fa5s.search", color=QColor(*color.as_rgb_tuple())))


class SearchROIButton(QtViewerPushButton):
    def __init__(self, settings: BaseSettings):
        super().__init__("fa5s.search", settings)
        self.setToolTip("Search component")



from abc import ABC
from typing import Callable, List, Optional, Tuple, Union

from PartSegCore.algorithm_describe_base import AlgorithmDescribeBase, AlgorithmProperty
from PartSegCore.roi_info import ROIInfo
from PartSegImage import Image


class TransformBase(AlgorithmDescribeBase, ABC):
    @classmethod
    def transform(
        cls,
        image: Image,
        roi_info: ROIInfo,
        arguments: dict,
        callback_function: Optional[Callable[[str, int], None]] = None,
    ) -> Tuple[Image, Optional[ROIInfo]]:
        raise NotImplementedError

    @classmethod
    def get_fields_per_dimension(cls, component_list: List[str]) -> List[Union[str, AlgorithmProperty]]:
        raise NotImplementedError

    @classmethod
    def calculate_initial(cls, image: Image):
        raise NotImplementedError



import dataclasses
import typing
from collections import defaultdict
from copy import copy, deepcopy
from os import path

import numpy as np
from qtpy.QtCore import Signal, Slot
from qtpy.QtWidgets import QMessageBox

from PartSeg.common_backend.base_settings import BaseSettings
from PartSegCore.algorithm_describe_base import ROIExtractionProfile
from PartSegCore.io_utils import PointsInfo
from PartSegCore.mask.io_functions import MaskProjectTuple, load_metadata
from PartSegCore.project_info import HistoryElement, HistoryProblem
from PartSegCore.roi_info import ROIInfo
from PartSegCore.segmentation.algorithm_base import ROIExtractionResult
from PartSegImage import Image
from PartSegImage.image import minimal_dtype, reduce_array


class StackSettings(BaseSettings):
    load_metadata = staticmethod(load_metadata)
    components_change_list = Signal([int, list])
    save_locations_keys: typing.ClassVar[typing.List[str]] = [
        "save_batch",
        "save_components_directory",
        "save_segmentation_directory",
        "open_segmentation_directory",
        "load_image_directory",
        "batch_directory",
        "multiple_open_directory",
    ]

    def __init__(self, json_path, profile_name="default"):
        super().__init__(json_path, profile_name)
        self.chosen_components_widget = None
        self.keep_chosen_components = False
        self.components_parameters_dict: typing.Dict[int, ROIExtractionProfile] = {}

    def set_segmentation_result(self, result: ROIExtractionResult):
        if (
            result.file_path is not None and result.file_path and result.file_path != self.image.file_path
        ):  # pragma: no cover
            if self._parent is not None:
                # TODO change to non disrupting popup
                QMessageBox().warning(
                    self._parent, "Result file bug", "It looks like one try to set ROI form another file."
                )
            return
        if self._parent and len(result.roi_info.bound_info) == 0:  # pragma: no cover
            QMessageBox.information(
                self._parent,
                "No result",
                "Segmentation contains no component, check parameters, especially chosen channel.",
            )
        if result.info_text and self._parent is not None:  # pragma: no cover
            QMessageBox().information(self._parent, "Algorithm info", result.info_text)
        parameters_dict = defaultdict(lambda: deepcopy(result.parameters))
        self._additional_layers = result.additional_layers
        self.additional_layers_changed.emit()
        self.last_executed_algorithm = result.parameters.algorithm
        self.set_algorithm(f"algorithms.{result.parameters.algorithm}", result.parameters.values)
        self._set_roi_info(result.roi_info, True, [], parameters_dict)
        if result.points is not None:
            self.points = result.points

    @Slot(int)
    def set_keep_chosen_components(self, val: bool):
        self.keep_chosen_components = val

    def file_save_name(self):
        return path.splitext(path.basename(self.image.file_path))[0]

    def get_file_names_for_save_result(self, dir_path):
        components = self.chosen_components()
        file_name = self.file_save_name()
        res = []
        for i in components:
            res.extend(
                (
                    path.join(dir_path, f"{file_name}_component{i}.tif"),
                    path.join(dir_path, f"{file_name}_component{i}_mask.tif"),
                )
            )

        return res

    def chosen_components(self) -> typing.List[int]:
        """
        Needs instance of :py:class:`PartSeg.segmentation_mask.main_window.ChosenComponents` on variable
        Py:attr:`chosen_components_widget` (or something implementing its interface)

        :return: list of chosen components
        """
        if self.chosen_components_widget is not None:
            return sorted(self.chosen_components_widget.get_chosen())

        raise RuntimeError("chosen_components_widget do not initialized")  # pragma: no cover

    def component_is_chosen(self, val: int) -> bool:
        """
        Needs instance of :py:class:`PartSeg.segmentation_mask.main_window.ChosenComponents` on variable
        Py:attr:`chosen_components_widget` (or something implementing its interface)

        :return: if given component is selected
        """
        if self.chosen_components_widget is not None:
            return self.chosen_components_widget.get_state(val)

        raise RuntimeError("chosen_components_widget do not idealized")  # pragma: no cover

    def components_mask(self) -> np.ndarray:
        """
        Needs instance of :py:class:`PartSeg.segmentation_mask.main_window.ChosenComponents` on variable
        Py:attr:`chosen_components_widget` (or something implementing its interface)

        :return: boolean mask if component is selected
        """
        if self.chosen_components_widget is not None:
            return self.chosen_components_widget.get_mask()

        raise RuntimeError("chosen_components_widget do not initialized")  # pragma: no cover

    def get_project_info(self) -> MaskProjectTuple:
        """
        Get all information about current project

        :return: object with all project details.
        :rtype: MaskProjectTuple
        """
        return MaskProjectTuple(
            file_path=self.image.file_path,
            image=self.image.substitute(),
            mask=self.mask,
            roi_info=self.roi_info,
            selected_components=self.chosen_components(),
            roi_extraction_parameters=copy(self.components_parameters_dict),
            history=self.history[: self.history_index + 1],
            points=self.points,
        )

    def set_project_info(self, data: typing.Union[MaskProjectTuple, PointsInfo]):
        """signals = self.signalsBlocked()
        if data.segmentation is not None:
            self.blockSignals(True)"""

        if isinstance(data, PointsInfo):
            self.points = data.points
            return
        if isinstance(data.image, Image) and (
            self.image_path != data.image.file_path or self.image_shape != data.image.shape
        ):
            # This line clean segmentation also
            self.image = data.image
        state = self.get_project_info()
        # TODO Remove repetition this and set_segmentation code

        components = sorted(data.roi_info.bound_info)
        for i in components:
            _skip = data.roi_extraction_parameters[int(i)]
        self.mask = data.mask
        if self.keep_chosen_components:
            if not self.compare_history(data.history) and self.chosen_components():
                raise HistoryProblem("Incompatible history")
            state2 = self.transform_state(
                state,
                data.roi_info,
                data.roi_extraction_parameters,
                data.selected_components,
                self.keep_chosen_components,
            )
            self.chosen_components_widget.set_chose(
                sorted(state2.roi_extraction_parameters.keys()), state2.selected_components
            )
            self.roi = state2.roi_info
            self.components_parameters_dict = state2.roi_extraction_parameters
        else:
            self.set_history(data.history)
            self.chosen_components_widget.set_chose(
                sorted(data.roi_extraction_parameters.keys()), data.selected_components
            )
            self.roi = data.roi_info
            self.components_parameters_dict = data.roi_extraction_parameters

    @staticmethod
    def _clip_data_array(mask_array: np.ndarray, data_array: np.ndarray) -> np.ndarray:
        """
        From `data_array` fet only information masked with `mask_array`

        :param np.ndarray mask_array: masking array
        :param np.ndarray data_array: data array
        :return: array with selected data
        :rtype: np.ndarray
        """
        res = np.copy(data_array)
        res[mask_array == 0] = 0
        return res

    @classmethod
    def transform_state(
        cls,
        state: MaskProjectTuple,
        new_roi_info: ROIInfo,
        new_roi_extraction_parameters: typing.Dict[int, typing.Optional[ROIExtractionProfile]],
        list_of_components: typing.List[int],
        save_chosen: bool = True,
    ) -> MaskProjectTuple:
        """

        :param MaskProjectTuple state: state to be transformed
        :param ROIInfo new_roi_info: roi description
        :param typing.Dict[int, typing.Optional[ROIExtractionProfile]] new_roi_extraction_parameters:
            Parameters used to extract roi
        :param typing.List[int] list_of_components: list of components from new_roi which should be selected
        :param bool save_chosen: if save currently selected components
        :return: new state
        """

        # TODO Refactor
        if not save_chosen or state.roi_info.roi is None or len(state.selected_components) == 0:
            return dataclasses.replace(
                state,
                roi_info=new_roi_info,
                selected_components=list_of_components,
                roi_extraction_parameters={i: new_roi_extraction_parameters[i] for i in new_roi_info.bound_info},
            )
        if list_of_components is None:
            list_of_components = []
        if new_roi_extraction_parameters is None:
            new_roi_extraction_parameters = defaultdict(lambda: None)
        segmentation_count = len(state.roi_info.bound_info)
        new_segmentation_count = len(new_roi_info.bound_info)
        segmentation_dtype = minimal_dtype(segmentation_count + new_segmentation_count)
        roi_base = reduce_array(state.roi_info.roi, state.selected_components, dtype=segmentation_dtype)
        annotation_base = {
            i: state.roi_info.annotations.get(x) for i, x in enumerate(state.selected_components, start=1)
        }
        alternative_base = {
            name: cls._clip_data_array(roi_base, array) for name, array in state.roi_info.alternative.items()
        }
        components_parameters_dict = {
            i: state.roi_extraction_parameters[val] for i, val in enumerate(sorted(state.selected_components), 1)
        }

        base_chose = list(annotation_base.keys())

        if new_segmentation_count == 0:
            return dataclasses.replace(
                state,
                roi_info=ROIInfo(roi=roi_base, annotations=annotation_base, alternative=alternative_base),
                selected_components=base_chose,
                roi_extraction_parameters=components_parameters_dict,
            )

        new_segmentation = np.copy(new_roi_info.roi)
        new_segmentation[roi_base > 0] = 0
        left_component_list = np.unique(new_segmentation.flat)
        if left_component_list[0] == 0:
            left_component_list = left_component_list[1:]
        new_segmentation = reduce_array(new_segmentation, dtype=segmentation_dtype)
        roi_base[new_segmentation > 0] = new_segmentation[new_segmentation > 0] + len(base_chose)
        for name, array in new_roi_info.alternative.items():
            if name in alternative_base:
                alternative_base[name][new_segmentation > 0] = array[new_segmentation > 0]
            else:
                alternative_base[name] = cls._clip_data_array(new_segmentation, array)
        for i, el in enumerate(left_component_list, start=len(base_chose) + 1):
            annotation_base[i] = new_roi_info.annotations.get(el)
            if el in list_of_components:
                base_chose.append(i)
            components_parameters_dict[i] = new_roi_extraction_parameters[el]

        roi_info = ROIInfo(roi=roi_base, annotations=annotation_base, alternative=alternative_base)

        return dataclasses.replace(
            state,
            roi_info=roi_info,
            selected_components=base_chose,
            roi_extraction_parameters=components_parameters_dict,
        )

    def compare_history(self, history: typing.List[HistoryElement]):
        # TODO check dict comparison
        if len(history) != self.history_size():
            return False
        return not any(
            el2.mask_property != el1.mask_property or el2.roi_extraction_parameters != el1.roi_extraction_parameters
            for el1, el2 in zip(self.history, history)
        )

    def _set_roi_info(
        self, new_roi_info: ROIInfo, save_chosen=True, list_of_components=None, segmentation_parameters=None
    ):
        if list_of_components is None:
            list_of_components = []
        if segmentation_parameters is None:
            segmentation_parameters = defaultdict(lambda: None)
        state = self.get_project_info()
        try:
            new_roi_info = new_roi_info.fit_to_image(self.image)
        except ValueError as e:
            raise ValueError("ROI do not fit to image") from e
        if save_chosen:
            state2 = self.transform_state(state, new_roi_info, segmentation_parameters, list_of_components, save_chosen)
            self.chosen_components_widget.set_chose(
                sorted(state2.roi_extraction_parameters.keys()), state2.selected_components
            )
            self.roi = state2.roi_info
            self.components_parameters_dict = state2.roi_extraction_parameters
        else:
            selected_parameters = {i: segmentation_parameters[i] for i in new_roi_info.bound_info}

            self.chosen_components_widget.set_chose(sorted(selected_parameters.keys()), list_of_components)
            self.roi = new_roi_info
            self.components_parameters_dict = segmentation_parameters


def get_mask(
    segmentation: typing.Optional[np.ndarray], mask: typing.Optional[np.ndarray], selected: typing.List[int]
) -> np.ndarray:
    """
    Calculate mask base on segmentation, current mask and list of chosen components.
    Its exclude selected components from mask.

    :param typing.Optional[np.ndarray] segmentation: segmentation array
    :param typing.Optional[np.ndarray] mask: current mask
    :param typing.List[int] selected: list of selected components which should be masked as non segmentation area
    :return: new mask
    :rtype: typing.Optional[np.ndarray]
    """
    if segmentation is None or len(selected) == 0:
        return mask
    segmentation = reduce_array(segmentation, selected)
    if mask is None:
        resp = np.ones(segmentation.shape, dtype=np.uint8)
    else:
        resp = np.copy(mask)
    resp[segmentation > 0] = 0
    return resp



"""
This module contains roi_extraction algorithms
"""

import os
import sys

from PartSegCore.segmentation.algorithm_base import ROIExtractionAlgorithm, ROIExtractionResult
from PartSegCore.segmentation.noise_filtering import NoiseFilteringBase
from PartSegCore.segmentation.restartable_segmentation_algorithms import RestartableAlgorithm
from PartSegCore.segmentation.segmentation_algorithm import StackAlgorithm
from PartSegCore.segmentation.threshold import BaseThreshold
from PartSegCore.segmentation.watershed import BaseWatershed

__all__ = [
    "ROIExtractionAlgorithm",
    "BaseWatershed",
    "NoiseFilteringBase",
    "BaseThreshold",
    "RestartableAlgorithm",
    "ROIExtractionResult",
    "StackAlgorithm",
]


if os.path.basename(sys.argv[0]) in ["sphinx-build", "sphinx-build.exe"]:
    for el in __all__:
        globals()[el].__module__ = __name__



import warnings

from PartSegCore.algorithm_describe_base import AlgorithmSelection
from PartSegCore.segmentation.restartable_segmentation_algorithms import (
    BorderRim,
    LowerThresholdAlgorithm,
    LowerThresholdFlowAlgorithm,
    MaskDistanceSplit,
    OtsuSegment,
    RangeThresholdAlgorithm,
    UpperThresholdAlgorithm,
    UpperThresholdFlowAlgorithm,
)


class AnalysisAlgorithmSelection(
    AlgorithmSelection,
    class_methods=["support_time", "support_z"],
    methods=["set_image", "set_mask", "get_info_text", "calculation_run"],
):
    """Register for segmentation method visible in PartSeg ROI Analysis."""


AnalysisAlgorithmSelection.register(LowerThresholdAlgorithm)
AnalysisAlgorithmSelection.register(UpperThresholdAlgorithm)
AnalysisAlgorithmSelection.register(RangeThresholdAlgorithm)
AnalysisAlgorithmSelection.register(LowerThresholdFlowAlgorithm, old_names=["Lower threshold flow"])
AnalysisAlgorithmSelection.register(UpperThresholdFlowAlgorithm, old_names=["Upper threshold flow"])
AnalysisAlgorithmSelection.register(OtsuSegment)
AnalysisAlgorithmSelection.register(BorderRim)
AnalysisAlgorithmSelection.register(MaskDistanceSplit, old_names=["Split Mask on Part"])


def __getattr__(name):  # pragma: no cover
    if name == "analysis_algorithm_dict":
        warnings.warn(
            "analysis_algorithm_dict is deprecated. Please use AnalysisAlgorithmSelection instead",
            category=FutureWarning,
            stacklevel=2,
        )
        return AnalysisAlgorithmSelection.__register__
    raise AttributeError(f"module {__name__} has no attribute {name}")



from PartSeg._launcher.main_window import PartSegGUILauncher
from PartSeg.plugins.napari_widgets.algorithm_widgets import (
    BorderSmooth,
    ConnectedComponents,
    DoubleThreshold,
    NoiseFilter,
    SplitCoreObjects,
    Threshold,
    Watershed,
)
from PartSeg.plugins.napari_widgets.colormap_control import ImageColormap
from PartSeg.plugins.napari_widgets.copy_labels import CopyLabelsWidget
from PartSeg.plugins.napari_widgets.lables_control import LabelSelector
from PartSeg.plugins.napari_widgets.mask_create_widget import MaskCreate
from PartSeg.plugins.napari_widgets.roi_extraction_algorithms import ROIAnalysisExtraction, ROIMaskExtraction
from PartSeg.plugins.napari_widgets.search_label_widget import SearchLabel

__all__ = (
    "PartSegGUILauncher",
    "BorderSmooth",
    "ConnectedComponents",
    "CopyLabelsWidget",
    "DoubleThreshold",
    "NoiseFilter",
    "SplitCoreObjects",
    "Threshold",
    "Watershed",
    "ImageColormap",
    "LabelSelector",
    "MaskCreate",
    "ROIAnalysisExtraction",
    "ROIMaskExtraction",
    "SearchLabel",
)



import operator
from abc import ABC
from itertools import product
from math import ceil
from typing import Callable, Optional

import numpy as np
import SimpleITK as sitk
from local_migrator import register_class, rename_key
from pydantic import Field

from PartSegCore.convex_fill import convex_fill
from PartSegCore.project_info import AdditionalLayerDescription
from PartSegCore.segmentation.algorithm_base import ROIExtractionAlgorithm, ROIExtractionResult
from PartSegCore.segmentation.border_smoothing import NoneSmoothing, OpeningSmoothing, SmoothAlgorithmSelection
from PartSegCore.segmentation.noise_filtering import NoiseFilterSelection
from PartSegCore.segmentation.threshold import BaseThreshold, DoubleThresholdSelection, ThresholdSelection
from PartSegCore.segmentation.utils import close_small_holes
from PartSegCore.segmentation.watershed import BaseWatershed, WatershedSelection
from PartSegCore.utils import BaseModel, bisect
from PartSegImage import Channel
from PartSegImage.image import minimal_dtype


class StackAlgorithm(ROIExtractionAlgorithm, ABC):
    def __init__(self):
        super().__init__()
        self.channel_num = 0

    @classmethod
    def support_time(cls):
        return False

    @classmethod
    def support_z(cls):
        return True

    def get_noise_filtered_channel(self, channel_idx, noise_removal):
        channel = self.get_channel(channel_idx)
        return NoiseFilterSelection[noise_removal.name].noise_filter(channel, self.image.spacing, noise_removal.values)


@register_class(version="0.0.1", migrations=[("0.0.1", rename_key("noise_removal", "noise_filtering", optional=True))])
class ThresholdPreviewParameters(BaseModel):
    channel: Channel = 0
    noise_filtering: NoiseFilterSelection = Field(NoiseFilterSelection.get_default(), title="Filter")
    threshold: int = Field(1000, ge=0, le=10**6)


class ThresholdPreview(StackAlgorithm):
    __argument_class__ = ThresholdPreviewParameters

    new_parameters: ThresholdPreviewParameters

    @classmethod
    def get_name(cls):
        return "Only Threshold"

    def calculation_run(self, report_fun) -> ROIExtractionResult:
        image = self.get_noise_filtered_channel(self.new_parameters.channel, self.new_parameters.noise_filtering)
        report_fun("threshold", 0)
        res = (image > self.new_parameters.threshold).astype(np.uint8)
        report_fun("mask", 1)
        if self.mask is not None:
            res[self.mask == 0] = 0
        report_fun("result", 2)
        val = ROIExtractionResult(
            roi=res,
            parameters=self.get_segmentation_profile(),
            additional_layers={"denoised image": AdditionalLayerDescription(layer_type="image", data=image)},
        )
        report_fun("return", 4)
        return val

    def get_info_text(self):
        return f"Threshold: {self.new_parameters.threshold}"

    @staticmethod
    def get_steps_num():
        return 3


def _migrate_smooth_border(dkt: dict):
    if isinstance(dkt["smooth_border"], bool):
        dkt = dkt.copy()
        if dkt["smooth_border"] and "smooth_border_radius" in dkt:
            dkt["smooth_border"] = SmoothAlgorithmSelection(
                name=OpeningSmoothing.get_name(),
                values=OpeningSmoothing.__argument_class__(smooth_border_radius=dkt.pop("smooth_border_radius")),
            )
        else:
            dkt["smooth_border"] = SmoothAlgorithmSelection(
                name=NoneSmoothing.get_name(), values=NoneSmoothing.__argument_class__()
            )
        if "smooth_border_radius" in dkt:
            del dkt["smooth_border_radius"]
    return dkt


@register_class(
    version="0.0.2",
    migrations=[
        ("0.0.1", _migrate_smooth_border),
        ("0.0.2", rename_key("noise_removal", "noise_filtering", optional=True)),
    ],
)
class BaseThresholdAlgorithmParameters(BaseModel):
    channel: Channel = 0
    noise_filtering: NoiseFilterSelection = Field(NoiseFilterSelection.get_default(), title="Filter")
    threshold: ThresholdSelection = Field(ThresholdSelection.get_default(), title="Threshold")
    close_holes: bool = Field(True, title="Fill holes")
    close_holes_size: int = Field(200, title="Maximum holes size (px)", ge=0, le=10**5)
    smooth_border: SmoothAlgorithmSelection = Field(SmoothAlgorithmSelection.get_default(), title="Smooth borders")
    side_connection: bool = Field(
        False,
        title="Side by Side connections",
        description="During calculation of connected components includes only side by side connected pixels",
    )
    minimum_size: int = Field(8000, ge=20, le=10**6)
    use_convex: int = Field(False, title="Use convex hull")


class BaseThresholdAlgorithm(StackAlgorithm, ABC):
    __argument_class__ = BaseThresholdAlgorithmParameters

    new_parameters: BaseThresholdAlgorithmParameters

    def __init__(self):
        super().__init__()
        self.sizes = [0]

    def get_info_text(self):
        if len(self.sizes) > 1:
            return f"ROI sizes: {', '.join(map(str, self.sizes[1:]))}"
        return ""


class BaseSingleThresholdAlgorithm(BaseThresholdAlgorithm, ABC):
    def __init__(self):
        super().__init__()
        self.base_sizes = [0]

    @staticmethod
    def get_steps_num():
        return 7

    def _threshold_image(self, image: np.ndarray) -> np.ndarray:
        raise NotImplementedError

    def _threshold_and_exclude(self, image, report_fun):
        raise NotImplementedError

    def _segmentation_calculation(self, binary_image):
        return sitk.GetArrayFromImage(
            sitk.RelabelComponent(
                sitk.ConnectedComponent(sitk.GetImageFromArray(binary_image), not self.new_parameters.side_connection),
                20,
            )
        )

    def calculation_run(self, report_fun):
        report_fun("Noise removal", 0)
        image = self.get_noise_filtered_channel(self.new_parameters.channel, self.new_parameters.noise_filtering)
        mask = self._threshold_and_exclude(image, report_fun)
        if self.new_parameters.close_holes:
            report_fun("Filing holes", 3)
            mask = close_small_holes(mask, self.new_parameters.close_holes_size)
        report_fun("Smooth border", 4)
        segmentation = SmoothAlgorithmSelection[self.new_parameters.smooth_border.name].smooth(
            mask, self.new_parameters.smooth_border.values
        )

        report_fun("Components calculating", 5)
        segmentation = self._segmentation_calculation(segmentation)

        self.base_sizes = np.bincount(segmentation.flat)
        ind = bisect(self.base_sizes[1:], self.new_parameters.minimum_size, lambda x, y: x > y)
        resp = np.copy(segmentation)
        resp[resp > ind] = 0

        if len(self.base_sizes) == 1:
            info_text = "Please check the threshold parameter. There is no object bigger than 20 voxels."
        elif ind == 0:
            info_text = f"Please check the minimum size parameter. The biggest element has size {self.base_sizes[1]}"
        else:
            info_text = ""
        self.sizes = self.base_sizes[: ind + 1]
        if self.new_parameters.use_convex:
            report_fun("convex hull", 6)
            resp = convex_fill(resp)
            self.sizes = np.bincount(resp.flat)

        report_fun("Calculation done", 7)
        return ROIExtractionResult(
            roi=self.image.fit_array_to_image(resp),
            parameters=self.get_segmentation_profile(),
            additional_layers={
                "denoised image": AdditionalLayerDescription(data=image, layer_type="image"),
                "no size filtering": AdditionalLayerDescription(data=self.segmentation, layer_type="labels"),
            },
            info_text=info_text,
            roi_annotation={i: {"voxels": v} for i, v in enumerate(self.sizes[1:], start=1)},
        )

    def get_info_text(self):
        base_text = super().get_info_text()
        base_sizes = self.base_sizes[: self.sizes.size]
        if np.any(base_sizes != self.sizes):
            base_text += "\nBase ROI sizes " + ", ".join(map(str, base_sizes))
        return base_text


class MorphologicalWatershed(BaseSingleThresholdAlgorithm):
    def __init__(self):
        super().__init__()
        self.base_sizes = [0]

    @classmethod
    def get_name(cls):
        return "Morphological Watersheed"

    def _threshold_and_exclude(self, image, report_fun):
        report_fun("Threshold calculation", 1)
        threshold_algorithm: BaseThreshold = ThresholdSelection[self.new_parameters.threshold.name]
        mask, _thr_val = threshold_algorithm.calculate_mask(
            image, self.mask, self.new_parameters.threshold.values, operator.ge
        )
        report_fun("Threshold calculated", 2)
        return mask

    def _segmentation_calculation(self, binary_image):
        seg_image = sitk.GetImageFromArray(binary_image)
        distance_map = sitk.SignedMaurerDistanceMap(
            seg_image, insideIsPositive=False, squaredDistance=False, useImageSpacing=False
        )

        ws = sitk.MorphologicalWatershed(distance_map, markWatershedLine=False, level=1)
        return sitk.GetArrayFromImage(sitk.RelabelComponent(sitk.Mask(ws, sitk.Cast(seg_image, sitk.sitkUInt8)), 20))


class ThresholdAlgorithm(BaseSingleThresholdAlgorithm):
    @classmethod
    def get_name(cls):
        return "Threshold"

    def _threshold_image(self, image: np.ndarray) -> Optional[np.ndarray]:
        return None

    def _threshold_and_exclude(self, image, report_fun):
        report_fun("Threshold calculation", 1)
        threshold_algorithm: BaseThreshold = ThresholdSelection[self.new_parameters.threshold.name]
        mask, _thr_val = threshold_algorithm.calculate_mask(
            image, self.mask, self.new_parameters.threshold.values, operator.ge
        )
        report_fun("Threshold calculated", 2)
        return mask


@register_class(version="0.0.1", migrations=[("0.0.1", rename_key("sprawl_type", "flow_type"))])
class ThresholdFlowAlgorithmParameters(BaseThresholdAlgorithmParameters):
    threshold: DoubleThresholdSelection = Field(DoubleThresholdSelection.get_default())
    flow_type: WatershedSelection = Field(WatershedSelection.get_default())


class ThresholdFlowAlgorithm(BaseThresholdAlgorithm):
    __argument_class__ = ThresholdFlowAlgorithmParameters

    new_parameters: ThresholdFlowAlgorithmParameters

    @classmethod
    def get_name(cls) -> str:
        return "Threshold Flow"

    def calculation_run(self, report_fun: Callable[[str, int], None]) -> ROIExtractionResult:
        report_fun("Noise removal", 0)
        noise_filtered = self.get_noise_filtered_channel(
            self.new_parameters.channel, self.new_parameters.noise_filtering
        )

        report_fun("Threshold apply", 1)
        mask, thr = DoubleThresholdSelection[self.new_parameters.threshold.name].calculate_mask(
            noise_filtered, self.mask, self.new_parameters.threshold.values, operator.ge
        )
        core_objects = np.array(mask == 2).astype(np.uint8)

        report_fun("Core components calculating", 2)
        core_objects = sitk.GetArrayFromImage(
            sitk.RelabelComponent(
                sitk.ConnectedComponent(sitk.GetImageFromArray(core_objects), not self.new_parameters.side_connection),
                20,
            )
        )
        self.base_sizes = np.bincount(core_objects.flat)
        ind = bisect(self.base_sizes[1:], self.new_parameters.minimum_size, lambda x, y: x > y)
        core_objects[core_objects > ind] = 0

        if self.new_parameters.close_holes:
            report_fun("Filing holes", 3)
            mask = close_small_holes(mask, self.new_parameters.close_holes_size)

        report_fun("Smooth border", 4)
        mask = SmoothAlgorithmSelection[self.new_parameters.smooth_border.name].smooth(
            mask, self.new_parameters.smooth_border.values
        )

        report_fun("Flow calculation", 5)
        sprawl_algorithm: BaseWatershed = WatershedSelection[self.new_parameters.flow_type.name]
        segmentation = sprawl_algorithm.sprawl(
            mask,
            core_objects,
            noise_filtered,
            ind,
            self.image.spacing,
            self.new_parameters.side_connection,
            operator.gt,
            self.new_parameters.flow_type.values,
            thr[1],
            thr[0],
        )
        if self.new_parameters.use_convex:
            report_fun("convex hull", 6)
            segmentation = convex_fill(segmentation)
        report_fun("Calculation done", 7)
        return ROIExtractionResult(
            roi=segmentation,
            parameters=self.get_segmentation_profile(),
            additional_layers={
                "denoised image": AdditionalLayerDescription(data=noise_filtered, layer_type="image"),
                "no size filtering": AdditionalLayerDescription(data=mask, layer_type="labels"),
            },
        )

    @staticmethod
    def get_steps_num():
        return 7


class AutoThresholdAlgorithmParams(BaseThresholdAlgorithmParameters):
    suggested_size: int = Field(200000, ge=0, le=10**6)


class AutoThresholdAlgorithm(BaseSingleThresholdAlgorithm):
    __argument_class__ = AutoThresholdAlgorithmParams
    new_parameters: AutoThresholdAlgorithmParams

    @classmethod
    def get_name(cls):
        return "Auto Threshold"

    def _threshold_image(self, image: np.ndarray) -> np.ndarray:
        sitk_image = sitk.GetImageFromArray(image)
        sitk_mask = sitk.ThresholdMaximumConnectedComponents(sitk_image, self.new_parameters.suggested_size)
        # TODO what exactly it returns. Maybe it is already segmented.
        mask = sitk.GetArrayFromImage(sitk_mask)
        min_val = np.min(image[mask > 0])
        threshold_algorithm: BaseThreshold = ThresholdSelection[self.new_parameters.threshold.name]
        mask2, thr_val = threshold_algorithm.calculate_mask(
            image, None, self.new_parameters.threshold.values, operator.le
        )
        return mask if thr_val < min_val else mask2

    def _threshold_and_exclude(self, image, report_fun):
        if self.mask is not None:
            report_fun("Components exclusion apply", 1)
            image[self.mask == 0] = 0
        report_fun("Threshold calculation", 2)
        return self._threshold_image(image)


class CellFromNucleusFlowParameters(BaseModel):
    nucleus_channel: Channel = Field(0, title="Nucleus Channel")
    nucleus_noise_filtering: NoiseFilterSelection = Field(NoiseFilterSelection.get_default(), title="Filter")
    nucleus_threshold: ThresholdSelection = Field(ThresholdSelection.get_default(), title="Threshold")
    cell_channel: Channel = Field(0, title="Cell Channel")
    cell_noise_filtering: NoiseFilterSelection = Field(NoiseFilterSelection.get_default(), title="Filter")
    cell_threshold: ThresholdSelection = Field(ThresholdSelection.get_default(), title="Threshold")
    flow_type: WatershedSelection = Field(WatershedSelection.get_default(), title="Flow type")
    close_holes: bool = Field(True, title="Fill holes")
    close_holes_size: int = Field(200, title="Maximum holes size (px)", ge=0, le=10**5)
    smooth_border: SmoothAlgorithmSelection = Field(SmoothAlgorithmSelection.get_default(), title="Smooth borders")
    side_connection: bool = Field(
        False,
        title="Side by Side connections",
        description="During calculation of connected components includes only side by side connected pixels",
    )
    minimum_size: int = Field(8000, ge=20, le=10**6)
    use_convex: int = Field(False, title="Use convex hull")


class CellFromNucleusFlow(StackAlgorithm):
    __argument_class__ = CellFromNucleusFlowParameters
    new_parameters: CellFromNucleusFlowParameters

    def __init__(self):
        super().__init__()
        self.found_nucleus = 0

    def calculation_run(self, report_fun: Callable[[str, int], None]) -> ROIExtractionResult:
        report_fun("Nucleus noise removal", 0)
        nucleus_channel = self.get_noise_filtered_channel(
            self.new_parameters.nucleus_channel, self.new_parameters.nucleus_noise_filtering
        )
        report_fun("Nucleus threshold apply", 1)
        nucleus_mask, _nucleus_thr = ThresholdSelection[self.new_parameters.nucleus_threshold.name].calculate_mask(
            nucleus_channel, self.mask, self.new_parameters.nucleus_threshold.values, operator.ge
        )
        report_fun("Nucleus calculate", 2)
        nucleus_objects = sitk.GetArrayFromImage(
            sitk.RelabelComponent(
                sitk.ConnectedComponent(sitk.GetImageFromArray(nucleus_mask), not self.new_parameters.side_connection),
                20,
            )
        )
        sizes = np.bincount(nucleus_objects.flat)
        ind = bisect(sizes[1:], self.new_parameters.minimum_size, lambda x, y: x > y)
        nucleus_objects[nucleus_objects > ind] = 0
        self.found_nucleus = ind
        report_fun("Cell noise removal", 3)
        cell_channel = self.get_noise_filtered_channel(
            self.new_parameters.cell_channel, self.new_parameters.cell_noise_filtering
        )
        report_fun("Cell threshold apply", 4)
        cell_mask, cell_thr = ThresholdSelection[self.new_parameters.cell_threshold.name].calculate_mask(
            cell_channel, self.mask, self.new_parameters.cell_threshold.values, operator.ge
        )

        report_fun("Flow calculation", 5)
        sprawl_algorithm: BaseWatershed = WatershedSelection[self.new_parameters.flow_type.name]
        mean_brightness = np.mean(cell_channel[cell_mask > 0])
        if mean_brightness < cell_thr:
            mean_brightness = cell_thr + 10
        segmentation = sprawl_algorithm.sprawl(
            cell_mask,
            nucleus_objects,
            cell_channel,
            ind,
            self.image.spacing,
            self.new_parameters.side_connection,
            operator.gt,
            self.new_parameters.flow_type.values,
            cell_thr,
            mean_brightness,
        )
        report_fun("Smooth border", 6)
        segmentation = SmoothAlgorithmSelection[self.new_parameters.smooth_border.name].smooth(
            segmentation, self.new_parameters.smooth_border.values
        )
        if self.new_parameters.use_convex:
            report_fun("convex hull", 7)
            segmentation = convex_fill(segmentation)
        report_fun("Calculation done", 8)
        return ROIExtractionResult(
            roi=segmentation,
            parameters=self.get_segmentation_profile(),
            additional_layers={
                "no size filtering": AdditionalLayerDescription(data=cell_mask, layer_type="labels"),
            },
        )

    def get_info_text(self):
        return f"Fond nucleus: {self.found_nucleus}"

    @staticmethod
    def get_steps_num():
        return 9

    @classmethod
    def get_name(cls) -> str:
        return "Cell from nucleus flow"


class SplitImageOnPartsParameters(BaseModel):
    side_length: int = Field(1000, ge=100, le=10**5)


class SplitImageOnParts(StackAlgorithm):
    __argument_class__ = SplitImageOnPartsParameters
    new_parameters: SplitImageOnPartsParameters

    def __init__(self):
        super().__init__()
        self._components_count = -1

    def calculation_run(self, report_fun: Callable[[str, int], None]) -> ROIExtractionResult:
        if self.image is None:  # pragma: no cover
            raise ValueError("No image")
        report_fun("Splitting image", 0)
        image = self.image
        size = self.new_parameters.side_length
        count_components = ceil(image.shape[-1] / size) * ceil(image.shape[-2] / size)
        dtype = minimal_dtype(count_components)
        mask = np.zeros(image.shape, dtype=dtype)
        x_step = ceil(image.shape[-1] / size)
        y_step = ceil(image.shape[-2] / size)
        for cnt, (i, j) in enumerate(product(range(x_step), range(y_step)), start=1):
            mask[..., j * size : (j + 1) * size, i * size : (i + 1) * size] = cnt

        report_fun("Done", 1)
        self._components_count = count_components
        return ROIExtractionResult(
            roi=mask,
            parameters=self.get_segmentation_profile(),
        )

    def get_info_text(self):
        return f"Split on {self._components_count} parts"

    @staticmethod
    def get_steps_num():
        return 2

    @classmethod
    def get_name(cls) -> str:
        return "Split image on parts"


final_algorithm_list = [
    ThresholdAlgorithm,
    ThresholdFlowAlgorithm,
    MorphologicalWatershed,
    ThresholdPreview,
    AutoThresholdAlgorithm,
    CellFromNucleusFlow,
    SplitImageOnParts,
]



"""
Code from https://github.com/napari/napari-console
"""

# pylint: disable=no-value-for-parameter,unexpected-keyword-arg
import re
import sys
from typing import TYPE_CHECKING

from ipykernel.connect import get_connection_file
from ipykernel.inprocess.ipkernel import InProcessInteractiveShell
from ipykernel.zmqshell import ZMQInteractiveShell
from IPython import get_ipython
from IPython.terminal.interactiveshell import TerminalInteractiveShell
from qtconsole.client import QtKernelClient
from qtconsole.inprocess import QtInProcessKernelManager
from qtconsole.rich_jupyter_widget import RichJupyterWidget
from qtpy.QtGui import QColor

if TYPE_CHECKING:  # pragma: no cover
    from PartSeg.common_gui.main_window import BaseMainWindow


def _empty_push(_):
    """Empty function for push"""


def str_to_rgb(arg):
    """Convert an rgb string 'rgb(x,y,z)' to a list of ints [x,y,z]."""
    return list(map(int, re.match(r"rgb\((\d+),\s*(\d+),\s*(\d+)\)", arg).groups()))


# set default asyncio policy to be compatible with tornado
# Tornado 6 (at least) is not compatible with the default
# asyncio implementation on Windows
# Pick the older SelectorEventLoopPolicy on Windows
# if the known-incompatible default policy is in use.
# FIXME: if/when tornado supports the defaults in asyncio,
# remove and bump tornado requirement for py38
# borrowed from ipykernel:  https://github.com/ipython/ipykernel/pull/456

if sys.platform.startswith("win"):
    import asyncio

    if (
        hasattr(asyncio, "WindowsProactorEventLoopPolicy")
        and hasattr(asyncio, "WindowsSelectorEventLoopPolicy")
        and type(asyncio.get_event_loop_policy()) is asyncio.WindowsProactorEventLoopPolicy  # pylint: disable=C0123
    ):
        # WindowsProactorEventLoopPolicy is not compatible with tornado 6
        # fallback to the pre-3.8 default of Selector
        asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())


class QtConsole(RichJupyterWidget):  # pylint: disable=abstract-method
    """Qt view for the console, an integrated iPython terminal in napari.
    Parameters
    ----------
    main_window : BaseMainWindow
        Parent widget.
    Attributes
    ----------
    kernel_client : qtconsole.inprocess.QtInProcessKernelClient,
                    qtconsole.client.QtKernelClient, or None
        Client for the kernel if it exists, None otherwise.
    shell : ipykernel.inprocess.ipkernel.InProcessInteractiveShell,
            ipykernel.zmqshell.ZMQInteractiveShell, or None.
        Shell for the kernel if it exists, None otherwise.
    """

    def __init__(self, main_window: "BaseMainWindow"):
        super().__init__()

        self.main_window = main_window
        self.syntax_style = "default"
        self.style_sheet = ""

        # Connect theme update
        user_variables = {"window": self.main_window, "settings": self.main_window.settings}

        # get current running instance or create new instance
        shell = get_ipython()

        if shell is None:
            # If there is no currently running instance create an in-process
            # kernel.
            kernel_manager = QtInProcessKernelManager()
            kernel_manager.start_kernel(show_banner=False)
            kernel_manager.kernel.gui = "qt"

            kernel_client = kernel_manager.client()
            kernel_client.start_channels()

            self.kernel_manager = kernel_manager
            self.kernel_client = kernel_client
            self.shell = kernel_manager.kernel.shell
            self.push = self.shell.push
        elif type(shell) is InProcessInteractiveShell:  # pylint: disable=unidiomatic-typecheck
            # If there is an existing running InProcessInteractiveShell
            # it is likely because multiple viewers have been launched from
            # the same process. In that case create a new kernel.
            # Connect existing kernel
            kernel_manager = QtInProcessKernelManager(kernel=shell.kernel)
            kernel_client = kernel_manager.client()

            self.kernel_manager = kernel_manager
            self.kernel_client = kernel_client
            self.shell = kernel_manager.kernel.shell
            self.push = self.shell.push
        elif isinstance(shell, TerminalInteractiveShell):
            # if launching from an ipython terminal then adding a console is
            # not supported. Instead users should use the ipython terminal for
            # the same functionality.
            self.kernel_client = None
            self.kernel_manager = None
            self.shell = None
            self.push = _empty_push

        elif isinstance(shell, ZMQInteractiveShell):
            # if launching from jupyter notebook, connect to the existing
            # kernel
            kernel_client = QtKernelClient(connection_file=get_connection_file())
            kernel_client.load_connection_file()
            kernel_client.start_channels()

            self.kernel_manager = None
            self.kernel_client = kernel_client
            self.shell = shell
            self.push = self.shell.push
        else:
            raise ValueError(f"ipython shell not recognized; got {type(shell)}")
        # Add any user variables
        user_variables = user_variables or {}
        self.push(user_variables)

        self.enable_calltips = False

    def _update_theme(self, event=None):
        """Update the napari GUI theme."""
        from napari.qt import get_stylesheet
        from napari.utils.theme import get_theme, template

        # qtconsole unfortunately won't inherit the parent stylesheet
        # so it needs to be directly set
        raw_stylesheet = get_stylesheet()
        # template and apply the primary stylesheet
        # (should probably be done by napari)
        theme = get_theme(self.viewer.theme)
        self.style_sheet = template(raw_stylesheet, **theme)

        # Set syntax styling and highlighting using theme
        self.syntax_style = theme["syntax_style"]
        bracket_color = QColor(*str_to_rgb(theme["highlight"]))
        self._bracket_matcher.format.setBackground(bracket_color)

    def closeEvent(self, event):
        """Clean up the integrated console in napari."""

        if self.kernel_client is not None:
            self.kernel_client.stop_channels()
        if self.kernel_manager is not None and self.kernel_manager.has_kernel:
            self.kernel_manager.shutdown_kernel()

        # RichJupyterWidget doesn't clean these up
        self._completion_widget.deleteLater()
        self._call_tip_widget.deleteLater()
        self.deleteLater()
        event.accept()



# pylint: disable=no-self-use

import sys
import typing
from collections import OrderedDict

import pytest

from PartSegCore.algorithm_describe_base import Register
from PartSegCore.class_generator import BaseSerializableClass, base_serialize_register

pytestmark = [
    pytest.mark.skipif(sys.version_info >= (3, 10), reason="class register is not compatible with python 3.10+"),
    pytest.mark.filterwarnings("ignore:BaseSerializableClass is deprecated"),
]

copy_register = Register()


def setup_module():
    """setup any state specific to the execution of the given module."""
    from copy import deepcopy

    from PartSegCore import class_generator

    global copy_register  # pylint: disable=W0603  # noqa: PLW0603
    copy_register = deepcopy(class_generator.base_serialize_register)


def teardown_module():
    """teardown any state that was previously setup with a setup_module
    method.
    """
    from PartSegCore import class_generator

    class_generator.base_serialize_register = copy_register


def empty(*_):
    """Empty function to avoid unused variable warning."""


def test_readonly():
    class Test1(BaseSerializableClass):
        field1: str
        field2: int
        field3: float

    val = Test1("a", 1, 0.7)
    with pytest.raises(AttributeError):
        val.field1 = "a"
    with pytest.raises(AttributeError):
        val.field2 = 1
    with pytest.raises(AttributeError):
        val.field3 = 1.1
    assert val.field1 == "a"
    assert val.field2 == 1
    assert val.field3 == 0.7
    base_serialize_register.clear()


def test_default_values():
    class Test2(BaseSerializableClass):
        field1: str
        field2: int = 3
        field3: float = 11.1

    val1 = Test2("a")
    assert val1.field1 == "a"
    assert val1.field2 == 3
    assert val1.field3 == 11.1

    val2 = Test2("b", 7)
    assert val2.field1 == "b"
    assert val2.field2 == 7
    assert val2.field3 == 11.1

    val3 = Test2("c", field3=7.7)
    assert val3.field1 == "c"
    assert val3.field2 == 3
    assert val3.field3 == 7.7

    val4 = Test2(field3=5.5, field1="d")
    assert val4.field1 == "d"
    assert val4.field2 == 3
    assert val4.field3 == 5.5
    base_serialize_register.clear()


def test_functions_save():
    class Test3(BaseSerializableClass):
        field1: str
        field2: int

        def __str__(self):
            return f"Test3({self.field1}, {self.field2})"

        def test1(self):
            return self.field2 * self.field1

    val = Test3("aa", 4)
    assert str(val) == "Test3(aa, 4)"
    assert val.test1() == "aaaaaaaa"
    base_serialize_register.clear()


def test_name_collision():
    class Test4(BaseSerializableClass):
        field1: str

    empty(Test4)
    with pytest.raises(ValueError, match="class already registered"):

        class Test4(BaseSerializableClass):  # pylint: disable=E0102
            field1: str
            field2: str

    base_serialize_register.clear()
    empty(Test4)


def test_subclasses():
    class Test4(BaseSerializableClass):
        field1: str

    class Test5(Test4):
        field2: int

    assert issubclass(Test5, BaseSerializableClass)
    val0 = Test4("c")
    empty(val0)
    val = Test5("a", "b")
    assert val.field1 == "a"
    assert val.field2 == "b"
    assert isinstance(val, Test4)
    assert issubclass(Test5, Test4)
    base_serialize_register.clear()


def test_typing():
    class Test(BaseSerializableClass):
        field1: typing.Union[str, float]
        field2: typing.Any
        field3: typing.Optional[int]
        field4: typing.List[str]

    val = Test("a", [1, 2, 3], 5, ["a", "b"])
    assert val.field1 == "a"
    assert val.field2 == [1, 2, 3]
    assert val.field3 == 5
    assert val.field4 == ["a", "b"]

    class MyClass:
        def __init__(self, f=1, k=2):
            self.f = f
            self.k = k

    class Test2(BaseSerializableClass):
        field1: typing.Union[str, float]
        field2: typing.Any
        field3: typing.Optional[int]
        field4: typing.List[str]
        field5: typing.Optional[MyClass]

    Test2("aa", 1, None, ["b", "c"], MyClass())
    base_serialize_register.clear()


def test_forward_ref():
    class Test(BaseSerializableClass):  # pylint: disable=W0612 # skipcq: PTC-W0065
        val: int
        child: typing.Optional["Test"] = None

    base_serialize_register.clear()


def test_generic_types():
    class Test1(BaseSerializableClass):
        list1: typing.Optional[int]
        list2: typing.Union[str, int]

    class Test2(BaseSerializableClass):
        list1: typing.List[int]
        list2: typing.List
        dict1: typing.Dict[str, int]
        dict2: typing.Dict

    empty(Test1, Test2)
    base_serialize_register.clear()


def test_post_init():
    class Test1(BaseSerializableClass):
        field1: str
        field2: int
        field3: str = None

        def __post_init__(self):
            self.field3 = self.field1 * self.field2

        def __str__(self):
            return f"{self.field1}, {self.field2}, {self.field3}"

    with pytest.raises(AttributeError):
        Test1("a", 3)

    class Test2(BaseSerializableClass):
        __readonly__ = False
        field1: str
        field2: int
        field3: str = None

        def __post_init__(self):
            self.field3 = self.field1 * self.field2

        def __str__(self):
            return f"{self.field1}, {self.field2}, {self.field3}"

    val = Test2("a", 3)
    assert val.field3 == "aaa"
    base_serialize_register.clear()


def test_functions():
    class Test1(BaseSerializableClass):
        field1: str
        field2: int
        field3: float

    class Test2(BaseSerializableClass):
        __readonly__ = False
        field1: str
        field2: int
        field3: float

    val1 = Test1("a", 1, 0.7)
    val2 = Test2("b", 2, 0.9)
    assert val1.as_tuple() == ("a", 1, 0.7)
    assert val2.as_tuple() == ("b", 2, 0.9)
    val2.field1 = "c"
    assert val2.as_tuple() == ("c", 2, 0.9)
    val3 = val1.replace_(field1="d")
    assert val1.as_tuple() == ("a", 1, 0.7)
    assert val3.as_tuple() == ("d", 1, 0.7)
    val4 = Test1("d", 1, 0.7)
    val5 = Test2("d", 1, 0.7)
    # __eq__ test
    assert val4 != val5
    assert val3 != val5
    assert val3 == val4

    assert OrderedDict([("field1", "d"), ("field2", 1), ("field3", 0.7)]) == val5.asdict()
    assert val4.asdict() == val5.asdict()

    val6 = Test1.make_(("a", 1, 0.7))
    assert isinstance(val6, Test1)
    assert val6 == val1

    val6 = Test1.make_({"field1": "a", "field2": 1, "field3": 0.7})
    assert isinstance(val6, Test1)
    assert val6 == val1
    base_serialize_register.clear()


def test_statistic_type():
    from PartSegCore.analysis.measurement_base import Leaf, Node

    empty(Leaf, Node)
    base_serialize_register.clear()



from qtpy.QtGui import QMouseEvent

try:
    from qtpy import QT5
except ImportError:  # pragma: no cover
    QT5 = True

if QT5:

    def get_mouse_x(event: QMouseEvent) -> float:
        return event.x()

    def get_mouse_y(event: QMouseEvent) -> float:
        return event.y()

else:

    def get_mouse_x(event: QMouseEvent) -> float:
        return event.position().x()

    def get_mouse_y(event: QMouseEvent) -> float:
        return event.position().x()



import dataclasses
import itertools
import os
import platform
from copy import deepcopy
from pathlib import Path

import numpy as np
import pytest

from PartSeg import state_store
from PartSegCore.algorithm_describe_base import ROIExtractionProfile
from PartSegCore.analysis import ProjectTuple, SegmentationPipeline, SegmentationPipelineElement
from PartSegCore.analysis.measurement_base import AreaType, MeasurementEntry, PerComponent
from PartSegCore.analysis.measurement_calculation import ComponentsNumber, MeasurementProfile, Volume
from PartSegCore.image_operations import RadiusType
from PartSegCore.mask.io_functions import MaskProjectTuple
from PartSegCore.mask_create import MaskProperty
from PartSegCore.project_info import HistoryElement
from PartSegCore.roi_info import ROIInfo
from PartSegCore.segmentation.restartable_segmentation_algorithms import BorderRim, LowerThresholdAlgorithm
from PartSegCore.segmentation.segmentation_algorithm import ThresholdAlgorithm
from PartSegImage import Image


@pytest.fixture(scope="module")
def data_test_dir():
    """Return path to directory with test data that need to be downloaded"""
    return Path(__file__).absolute().parent.parent.parent / "test_data"


@pytest.fixture(autouse=True)
def _mock_settings_path(tmp_path, monkeypatch):
    monkeypatch.setattr(state_store, "save_folder", str(tmp_path))


@pytest.fixture(scope="module")
def bundle_test_dir():
    """Return path to directory with test data bundled in napari"""
    return Path(os.path.join(os.path.dirname(__file__), "test_data"))


@pytest.fixture()
def image(tmp_path):
    data = np.zeros([20, 20, 20, 2], dtype=np.uint8)
    data[10:-1, 1:-1, 1:-1, 0] = 20
    data[1:10, 1:-1, 1:-1, 1] = 20
    data[1:-1, 1:5, 1:-1, 1] = 20
    data[1:-1, -5:-1, 1:-1, 1] = 20

    return Image(data, (10**-3, 10**-3, 10**-3), axes_order="ZYXC", file_path=str(tmp_path / "test.tiff"))


@pytest.fixture()
def image2(image, tmp_path):
    data = np.zeros([20, 20, 20, 1], dtype=np.uint8)
    data[10:-1, 1:-1, 1:-1, 0] = 20
    img = image.merge(Image(data, (10**-3, 10**-3, 10**-3), axes_order="ZYXC"), "C")
    img.file_path = str(tmp_path / "test2.tiff")
    return img


@pytest.fixture()
def image2d(tmp_path):
    data = np.zeros([20, 20], dtype=np.uint8)
    data[10:-1, 1:-1] = 20
    return Image(data, (10**-3, 10**-3), axes_order="YX", file_path=str(tmp_path / "test.tiff"))


@pytest.fixture()
def stack_image():
    data = np.zeros([20, 40, 40], dtype=np.uint8)
    for x, y in itertools.product([0, 20], repeat=2):
        data[1:-1, x + 2 : x + 18, y + 2 : y + 18] = 100
    for x, y in itertools.product([0, 20], repeat=2):
        data[3:-3, x + 4 : x + 16, y + 4 : y + 16] = 120
    for x, y in itertools.product([0, 20], repeat=2):
        data[5:-5, x + 6 : x + 14, y + 6 : y + 14] = 140

    return MaskProjectTuple("test_path", Image(data, (2, 1, 1), axes_order="ZYX", file_path="test_path"))


@pytest.fixture()
def algorithm_parameters():
    algorithm_parameters = {
        "algorithm_name": "Lower threshold",
        "values": {
            "threshold": {"name": "Manual", "values": {"threshold": 10}},
            "channel": 0,
            "noise_filtering": {"name": "None", "values": {}},
            "minimum_size": 1,
            "side_connection": False,
        },
    }
    return deepcopy(algorithm_parameters)


@pytest.fixture()
def roi_extraction_profile():
    return ROIExtractionProfile(
        name="test",
        algorithm=LowerThresholdAlgorithm.get_name(),
        values=LowerThresholdAlgorithm.get_default_values(),
    )


@pytest.fixture()
def mask_segmentation_parameters():
    return ROIExtractionProfile(
        name="",
        algorithm="Threshold",
        values={
            "channel": 0,
            "noise_filtering": {"name": "None", "values": {}},
            "threshold": {"name": "Manual", "values": {"threshold": 10}},
            "close_holes": False,
            "close_holes_size": 200,
            "smooth_border": {"name": "None", "values": {}},
            "side_connection": False,
            "minimum_size": 10000,
            "use_convex": False,
        },
    )


@pytest.fixture()
def stack_segmentation1(stack_image: MaskProjectTuple, mask_segmentation_parameters):
    data = np.zeros([20, 40, 40], dtype=np.uint8)
    for i, (x, y) in enumerate(itertools.product([0, 20], repeat=2), start=1):
        data[1:-1, x + 2 : x + 18, y + 2 : y + 18] = i
    data = ROIInfo(stack_image.image.fit_array_to_image(data))
    parameters = {i: deepcopy(mask_segmentation_parameters) for i in range(1, 5)}
    return dataclasses.replace(
        stack_image, roi_info=data, roi_extraction_parameters=parameters, selected_components=[1, 3]
    )


@pytest.fixture()
def analysis_segmentation(stack_image: MaskProjectTuple):
    data = np.zeros([20, 40, 40], dtype=np.uint8)
    for i, (x, y) in enumerate(itertools.product([0, 20], repeat=2), start=1):
        data[1:-1, x + 2 : x + 18, y + 2 : y + 18] = i
    data = ROIInfo(stack_image.image.fit_array_to_image(data))
    return ProjectTuple(file_path=stack_image.file_path, image=stack_image.image, roi_info=data)


@pytest.fixture()
def analysis_segmentation2(analysis_segmentation: ProjectTuple):
    mask = (analysis_segmentation.roi_info.roi > 0).astype(np.uint8)
    return dataclasses.replace(analysis_segmentation, mask=mask)


@pytest.fixture()
def stack_segmentation2(stack_image: MaskProjectTuple, mask_segmentation_parameters):
    data = np.zeros([20, 40, 40], dtype=np.uint8)
    for i, (x, y) in enumerate(itertools.product([0, 20], repeat=2), start=1):
        data[3:-3, x + 4 : x + 16, y + 4 : y + 16] = i
    data = ROIInfo(stack_image.image.fit_array_to_image(data))
    mask_segmentation_parameters.values.threshold.values.threshold = 110
    parameters = {i: deepcopy(mask_segmentation_parameters) for i in range(1, 5)}
    return dataclasses.replace(
        stack_image, roi_info=data, roi_extraction_parameters=parameters, selected_components=[1, 3]
    )


@pytest.fixture()
def mask_property():
    return MaskProperty.simple_mask()


@pytest.fixture()
def mask_property_non_default():
    return MaskProperty(
        dilate=RadiusType.R2D,
        dilate_radius=10,
        fill_holes=RadiusType.R3D,
        max_holes_size=10,
        save_components=True,
        clip_to_mask=True,
        reversed_mask=True,
    )


@pytest.fixture()
def measurement_profiles():
    statistics = [
        MeasurementEntry(
            name="Segmentation Volume",
            calculation_tree=Volume.get_starting_leaf().replace_(area=AreaType.ROI, per_component=PerComponent.No),
        ),
        MeasurementEntry(
            name="ROI Components Number",
            calculation_tree=ComponentsNumber.get_starting_leaf().replace_(
                area=AreaType.ROI, per_component=PerComponent.No
            ),
        ),
    ]
    statistics2 = [
        MeasurementEntry(
            name="Mask Volume",
            calculation_tree=Volume.get_starting_leaf().replace_(area=AreaType.Mask, per_component=PerComponent.No),
        ),
    ]
    return MeasurementProfile(name="statistic1", chosen_fields=statistics), MeasurementProfile(
        name="statistic2", chosen_fields=statistics + statistics2
    )


@pytest.fixture()
def border_rim_profile():
    return ROIExtractionProfile(
        name="border_profile", algorithm=BorderRim.get_name(), values=BorderRim.get_default_values()
    )


@pytest.fixture()
def lower_threshold_profile():
    return ROIExtractionProfile(
        name="lower_profile",
        algorithm=LowerThresholdAlgorithm.get_name(),
        values=LowerThresholdAlgorithm.get_default_values(),
    )


@pytest.fixture()
def mask_threshold_profile():
    return ROIExtractionProfile(
        name="mask_profile",
        algorithm=ThresholdAlgorithm.get_name(),
        values=ThresholdAlgorithm.get_default_values(),
    )


@pytest.fixture()
def sample_pipeline(border_rim_profile, lower_threshold_profile, mask_property):
    return SegmentationPipeline(
        name="sample_pipeline",
        segmentation=border_rim_profile,
        mask_history=[SegmentationPipelineElement(segmentation=lower_threshold_profile, mask_property=mask_property)],
    )


@pytest.fixture()
def sample_pipeline2(border_rim_profile, lower_threshold_profile, mask_property):
    return SegmentationPipeline(
        name="sample_pipeline2",
        segmentation=lower_threshold_profile,
        mask_history=[SegmentationPipelineElement(segmentation=border_rim_profile, mask_property=mask_property)],
    )


@pytest.fixture()
def history_element(image, lower_threshold_profile):
    roi = np.zeros(image.shape, dtype=np.uint8)
    roi[0, 2:10] = 1
    roi[0, 10:-2] = 2

    return HistoryElement.create(
        roi_info=ROIInfo(roi),
        mask=None,
        roi_extraction_parameters={
            "algorithm_name": lower_threshold_profile.name,
            "values": lower_threshold_profile.values,
        },
        mask_property=MaskProperty.simple_mask(),
    )


def pytest_collection_modifyitems(session, config, items):
    image_tests = [x for x in items if "PartSegImage" in str(x.fspath)]
    core_tests = [x for x in items if "PartSegCore" in str(x.fspath)]
    other_test = [x for x in items if "PartSegCore" not in str(x.fspath) and "PartSegImage" not in str(x.fspath)]
    items[:] = image_tests + core_tests + other_test


def pytest_runtest_setup(item):
    if platform.system() == "Windows" and any(item.iter_markers(name="windows_ci_skip")):
        pytest.skip("glBindFramebuffer with no OpenGL")
    if platform.system() == "Windows" and any(item.iter_markers(name="pyside_skip")):
        import qtpy

        if qtpy.API_NAME == "PySide2":
            pytest.skip("PySide2 problems")
